from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.utils.html import mark_safe

from coderedcms.settings import cr_settings


def get_protected_media_link(request, path, render_link=False):
    if render_link:
        return mark_safe(
            "<a href='{0}{1}'>{0}{1}</a>".format(
                request.build_absolute_uri('/')[:-1],
                path
            )
        )
    return "{0}{1}".format(request.build_absolute_uri('/')[:-1], path)


def uri_validator(possible_uri):
    validate = URLValidator()
    try:
        validate(possible_uri)
        return True
    except ValidationError:
        return False


def attempt_protected_media_value_conversion(request, value):
    try:
        if value.startswith(cr_settings['PROTECTED_MEDIA_URL']):
            new_value = get_protected_media_link(request, value)
            return new_value
    except AttributeError:
        pass

    return value


def fix_ical_datetime_format(dt_str):
    """
    ICAL generation gives timezones in the format of 2018-06-30T14:00:00-04:00.
    The Timezone offset -04:00 has a character not recognized by the timezone offset
    code (%z).  The being the colon in -04:00.  We need it to instead be -0400
    """
    if dt_str and ":" == dt_str[-3:-2]:
        dt_str = dt_str[:-3] + dt_str[-2:]
        return dt_str
    return dt_str



"""
HTML blocks are simple blocks used to represent common HTML elements,
with additional styling and attributes.

HTML blocks should NOT contain more sub-blocks or sub-streamfields.
They must be safe to nest within more robust "content blocks" without
creating recursion.
"""
import logging
from django.utils.translation import gettext_lazy as _
from wagtail.contrib.table_block.blocks import TableBlock as WagtailTableBlock
from wagtail.core import blocks
from wagtail.documents.blocks import DocumentChooserBlock
from wagtail.embeds.blocks import EmbedBlock
from wagtail.images.blocks import ImageChooserBlock

from .base_blocks import (
    BaseBlock,
    BaseLinkBlock,
    ButtonMixin,
    ClassifierTermChooserBlock,
    CoderedAdvTrackingSettings,
    LinkStructValue,
)


logger = logging.getLogger('coderedcms')


class ButtonBlock(ButtonMixin, BaseLinkBlock):
    """
    A link styled as a button.
    """
    class Meta:
        template = 'coderedcms/blocks/button_block.html'
        icon = 'fa-hand-pointer-o'
        label = _('Button Link')
        value_class = LinkStructValue


class DownloadBlock(ButtonMixin, BaseBlock):
    """
    Link to a file that can be downloaded.
    """
    automatic_download = blocks.BooleanBlock(
        required=False,
        label=_('Auto download'),
    )
    downloadable_file = DocumentChooserBlock(
        required=False,
        label=_('Document link'),
    )

    advsettings_class = CoderedAdvTrackingSettings

    class Meta:
        template = 'coderedcms/blocks/download_block.html'
        icon = 'download'
        label = _('Download')


class EmbedGoogleMapBlock(BaseBlock):
    """
    An embedded Google map in an <iframe>.
    """
    search = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Search query'),
        help_text=_('Address or search term used to find your location on the map.'),
    )
    map_title = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Map title'),
        help_text=_('Map title for screen readers, ex: "Map to Goodale Park"')
    )
    place_id = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Google place ID'),
        help_text=_('Requires API key to use place ID.')
    )
    map_zoom_level = blocks.IntegerBlock(
        required=False,
        default=14,
        label=_('Map zoom level'),
        help_text=_(
            "Requires API key to use zoom. 1: World, 5: Landmass/continent, 10: City, 15: Streets, 20: Buildings"  # noqa
        )
    )

    class Meta:
        template = 'coderedcms/blocks/google_map.html'
        icon = 'fa-map'
        label = _('Google Map')


class EmbedVideoBlock(BaseBlock):
    """
    Embedded media using stock wagtail functionality.
    """
    url = EmbedBlock(
        required=True,
        label=_('URL'),
        help_text=_('Link to a YouTube/Vimeo video, tweet, facebook post, etc.')
    )

    class Meta:
        template = 'coderedcms/blocks/embed_video_block.html'
        icon = 'media'
        label = _('Embed Media')


class H1Block(BaseBlock):
    """
    An <h1> heading.
    """
    text = blocks.CharBlock(
        max_length=255,
        label=_('Text'),
    )

    class Meta:
        template = 'coderedcms/blocks/h1_block.html'
        icon = 'fa-header'
        label = _('Heading 1')


class H2Block(BaseBlock):
    """
    An <h2> heading.
    """
    text = blocks.CharBlock(
        max_length=255,
        label=_('Text'),
    )

    class Meta:
        template = 'coderedcms/blocks/h2_block.html'
        icon = 'fa-header'
        label = _('Heading 2')


class H3Block(BaseBlock):
    """
    An <h3> heading.
    """
    text = blocks.CharBlock(
        max_length=255,
        label=_('Text'),
    )

    class Meta:
        template = 'coderedcms/blocks/h3_block.html'
        icon = 'fa-header'
        label = _('Heading 3')


class TableBlock(BaseBlock):
    table = WagtailTableBlock()

    class Meta:
        template = 'coderedcms/blocks/table_block.html'
        icon = 'fa-table'
        label = 'Table'


class ImageBlock(BaseBlock):
    """
    An <img>, by default styled responsively to fill its container.
    """
    image = ImageChooserBlock(
        label=_('Image'),
    )

    class Meta:
        template = 'coderedcms/blocks/image_block.html'
        icon = 'image'
        label = _('Image')


class ImageLinkBlock(BaseLinkBlock):
    """
    An <a> with an image inside it, instead of text.
    """
    image = ImageChooserBlock(
        label=_('Image'),
    )
    alt_text = blocks.CharBlock(
        max_length=255,
        required=True,
        help_text=_('Alternate text to show if the image doesn’t load'),
    )

    class Meta:
        template = 'coderedcms/blocks/image_link_block.html'
        icon = 'image'
        label = _('Image Link')
        value_class = LinkStructValue


class PageListBlock(BaseBlock):
    """
    Renders a preview of selected pages.
    """
    indexed_by = blocks.PageChooserBlock(
        required=True,
        label=_('Parent page'),
        help_text=_(
            "Show a preview of pages that are children of the selected page. Uses ordering specified in the page’s LAYOUT tab."  # noqa
        ),
    )
    classified_by = ClassifierTermChooserBlock(
        required=False,
        label=_('Classified as'),
        help_text=_('Only show pages that are classified with this term.')
    )
    show_preview = blocks.BooleanBlock(
        required=False,
        default=False,
        label=_('Show body preview'),
    )
    num_posts = blocks.IntegerBlock(
        default=3,
        label=_('Number of pages to show'),
    )

    class Meta:
        template = 'coderedcms/blocks/pagelist_block.html'
        icon = 'list-ul'
        label = _('Latest Pages')

    def get_context(self, value, parent_context=None):

        context = super().get_context(value, parent_context=parent_context)

        indexer = value['indexed_by'].specific
        # try to use the CoderedPage `get_index_children()`,
        # but fall back to get_children if this is a non-CoderedPage
        if hasattr(indexer, 'get_index_children'):
            pages = indexer.get_index_children()
            if value['classified_by']:
                try:
                    pages = pages.filter(classifier_terms=value['classified_by'])
                except AttributeError:
                    # `pages` is not a queryset, or is not a queryset of CoderedPage.
                    logger.warning(
                        "Tried to filter by ClassifierTerm in PageListBlock, but <%s.%s ('%s')>.get_index_children()  # noqadid not return a queryset or is not a queryset of CoderedPage models.",  # noqa
                        indexer._meta.app_label, indexer.__class__.__name__, indexer.title
                    )
        else:
            pages = indexer.get_children().live()

        context['pages'] = pages[:value['num_posts']]
        return context


class PagePreviewBlock(BaseBlock):
    """
    Renders a preview of a specific page.
    """
    page = blocks.PageChooserBlock(
        required=True,
        label=_('Page to preview'),
        help_text=_('Show a mini preview of the selected page.'),
    )

    class Meta:
        template = 'coderedcms/blocks/pagepreview_block.html'
        icon = 'doc-empty-inverse'
        label = _('Page Preview')


class QuoteBlock(BaseBlock):
    """
    A <blockquote>.
    """
    text = blocks.TextBlock(
        required=True,
        rows=4,
        label=_('Quote Text'),
    )
    author = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Author'),
    )

    class Meta:
        template = 'coderedcms/blocks/quote_block.html'
        icon = 'openquote'
        label = _('Quote')


class RichTextBlock(blocks.RichTextBlock):
    class Meta:
        template = 'coderedcms/blocks/rich_text_block.html'



import string
import random

from bs4 import BeautifulSoup
from django import template
from django.conf import settings
from django.forms import ClearableFileInput
from django.utils.html import mark_safe
from wagtail.core.models import Collection
from wagtail.images.models import Image

from coderedcms import utils, __version__
from coderedcms.blocks import CoderedAdvSettings
from coderedcms.forms import SearchForm
from coderedcms.models import Footer, Navbar
from coderedcms.settings import cr_settings, get_bootstrap_setting
from coderedcms.models.wagtailsettings_models import LayoutSettings

register = template.Library()


@register.filter
def is_advanced_setting(obj):
    return CoderedAdvSettings in (obj.__class__,) + obj.__class__.__bases__


@register.filter
def is_file_form(form):
    return any([isinstance(field.field.widget, ClearableFileInput) for field in form])


@register.simple_tag
def coderedcms_version():
    return __version__


@register.simple_tag
def generate_random_id():
    value = ''.join(random.choice(string.ascii_letters + string.digits) for n in range(20))
    return "cr-{}".format(value)


@register.simple_tag
def is_menu_item_dropdown(value):
    return \
        len(value.get('sub_links', [])) > 0 or \
        (
            value.get('show_child_links', False) and
            len(value.get('page', []).get_children().live()) > 0
        )


@register.simple_tag(takes_context=True)
def is_active_page(context, curr_page, other_page):
    if hasattr(curr_page, 'get_url') and hasattr(other_page, 'get_url'):
        curr_url = curr_page.get_url(context['request'])
        other_url = other_page.get_url(context['request'])
        return curr_url == other_url
    return False


@register.simple_tag
def get_pictures(collection_id):
    collection = Collection.objects.get(id=collection_id)
    return Image.objects.filter(collection=collection)


@register.simple_tag(takes_context=True)
def get_navbar_css(context):
    layout = LayoutSettings.for_request(context['request'])
    fixed = "fixed-top" if layout.navbar_fixed else ""
    return " ".join([
        fixed,
        layout.navbar_collapse_mode,
        layout.navbar_color_scheme,
        layout.navbar_format,
        layout.navbar_class
    ])


@register.simple_tag
def get_navbars():
    return Navbar.objects.all()


@register.simple_tag
def get_footers():
    return Footer.objects.all()


@register.simple_tag
def get_searchform(request=None):
    if request:
        return SearchForm(request.GET)
    return SearchForm()


@register.simple_tag
def get_pageform(page, request):
    return page.get_form(request)


@register.simple_tag
def process_form_cell(request, cell):
    if isinstance(cell, str) and cell.startswith(cr_settings['PROTECTED_MEDIA_URL']):
        return utils.get_protected_media_link(request, cell, render_link=True)
    if utils.uri_validator(str(cell)):
        return mark_safe("<a href='{0}'>{1}</a>".format(cell, cell))
    return cell


@register.filter
def codered_settings(value):
    return cr_settings.get(value, None)


@register.filter
def bootstrap_settings(value):
    return get_bootstrap_setting(value)


@register.filter
def django_settings(value):
    return getattr(settings, value)


@register.simple_tag
def query_update(querydict, key=None, value=None):
    """
    Alters querydict (request.GET) by updating/adding/removing key to value
    """
    get = querydict.copy()
    if key:
        if value:
            get[key] = value
        else:
            try:
                del(get[key])
            except KeyError:
                pass
    return get


@register.simple_tag
def render_iframe_from_embed(embed):
    soup = BeautifulSoup(embed.html, "html.parser")
    try:
        iframe_tags = soup.find('iframe')
        iframe_tags['title'] = embed.title
        return mark_safe(soup.prettify())
    except AttributeError:
        pass
    except TypeError:
        pass

    return mark_safe(embed.html)


@register.filter
def map_to_bootstrap_alert(message_tag):
    """
    Converts a message level to a bootstrap 4 alert class
    """
    message_to_alert_dict = {
        'debug': 'primary',
        'info': 'info',
        'success': 'success',
        'warning': 'warning',
        'error': 'danger'
    }

    try:
        return message_to_alert_dict[message_tag]
    except KeyError:
        return ''



from django import forms


class ColorPickerWidget(forms.TextInput):
    input_type = 'color'


class ClassifierSelectWidget(forms.CheckboxSelectMultiple):
    template_name = 'coderedcms/widgets/checkbox_classifiers.html'

    def optgroups(self, name, value, attrs=None):
        from coderedcms.models.snippet_models import Classifier
        classifiers = Classifier.objects.all().select_related()

        groups = []
        has_selected = False

        for index, classifier in enumerate(classifiers):
            subgroup = []
            group_name = classifier.name
            subindex = 0
            choices = []

            for term in classifier.terms.all():
                choices.append((term.pk, term.name))

            groups.append((group_name, subgroup, index))

            for subvalue, sublabel in choices:
                selected = (
                    str(subvalue) in value and
                    (not has_selected or self.allow_multiple_selected)
                )
                has_selected |= selected
                subgroup.append(self.create_option(
                    name, subvalue, sublabel, selected, index,
                    subindex=subindex, attrs=attrs,
                ))
                if subindex is not None:
                    subindex += 1
        return groups



import re

from django.template import engines
from django.test import TestCase


django_engine = engines['django']
html_id_re = re.compile(r"^[A-Za-z][A-Za-z0-9_:\.-]*$")


class TemplateTagTests(TestCase):
    def test_coderedcms_generate_random_id(self):
        count = 1000
        t = django_engine.from_string(
            "{% load coderedcms_tags %}{% generate_random_id as rid %}{{rid}}"
        )
        ids = set([])
        for i in range(count):
            ids.add(t.render())

        # ensure we are reasonably unique
        self.assertEqual(len(ids), count)

        # ensure ids are valid
        for i, value in enumerate(ids, start=1):
            self.assertTrue(
                html_id_re.match(value),
                'ID #%s "%s" did not match regex %r' % (i, value, html_id_re),
            )



"""
Snippets are for content that is re-usable in nature.
"""

from django.db import models
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from modelcluster.fields import ParentalKey
from modelcluster.models import ClusterableModel
from wagtail.admin.edit_handlers import (
    FieldPanel,
    InlinePanel,
    MultiFieldPanel,
    StreamFieldPanel)
from wagtail.core.fields import StreamField
from wagtail.core.models import Orderable
from wagtail.images.edit_handlers import ImageChooserPanel
from wagtail.snippets.models import register_snippet
from wagtail.images import get_image_model_string

from coderedcms.blocks import HTML_STREAMBLOCKS, LAYOUT_STREAMBLOCKS, NAVIGATION_STREAMBLOCKS
from coderedcms.settings import cr_settings


@register_snippet
class Carousel(ClusterableModel):
    """
    Model that represents a Carousel. Can be modified through the snippets UI.
    Selected through Page StreamField bodies by the CarouselSnippetChooser in
    snippet_choosers.py
    """
    class Meta:
        verbose_name = _('Carousel')

    name = models.CharField(
        max_length=255,
        verbose_name=_('Name'),
    )
    show_controls = models.BooleanField(
        default=True,
        verbose_name=_('Show controls'),
        help_text=_('Shows arrows on the left and right of the carousel to advance next or previous slides.'),  # noqa
    )
    show_indicators = models.BooleanField(
        default=True,
        verbose_name=_('Show indicators'),
        help_text=_('Shows small indicators at the bottom of the carousel based on the number of slides.'),  # noqa
    )
    animation = models.CharField(
        blank=True,
        max_length=20,
        choices=cr_settings['FRONTEND_CAROUSEL_FX_CHOICES'],
        default=cr_settings['FRONTEND_CAROUSEL_FX_DEFAULT'],
        verbose_name=_('Animation'),
        help_text=_('The animation when transitioning between slides.'),
    )

    panels = (
        [
            MultiFieldPanel(
                heading=_('Slider'),
                children=[
                    FieldPanel('name'),
                    FieldPanel('show_controls'),
                    FieldPanel('show_indicators'),
                    FieldPanel('animation'),
                ]
            ),
            InlinePanel('carousel_slides', label=_('Slides'))
        ]
    )

    def __str__(self):
        return self.name


class CarouselSlide(Orderable, models.Model):
    """
    Represents a slide for the Carousel model. Can be modified through the
    snippets UI.
    """
    class Meta:
        verbose_name = _('Carousel Slide')

    carousel = ParentalKey(
        Carousel,
        related_name='carousel_slides',
        verbose_name=_('Carousel'),
    )
    image = models.ForeignKey(
        get_image_model_string(),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name='+',
        verbose_name=_('Image'),
    )
    background_color = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Background color'),
        help_text=_('Hexadecimal, rgba, or CSS color notation (e.g. #ff0011)'),
    )
    custom_css_class = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Custom CSS class'),
    )
    custom_id = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Custom ID'),
    )

    content = StreamField(HTML_STREAMBLOCKS, blank=True)

    panels = (
        [
            ImageChooserPanel('image'),
            FieldPanel('background_color'),
            FieldPanel('custom_css_class'),
            FieldPanel('custom_id'),
            StreamFieldPanel('content'),
        ]
    )


@register_snippet
class Classifier(ClusterableModel):
    """
    Simple and generic model to organize/categorize/group pages.
    """
    class Meta:
        verbose_name = _('Classifier')
        verbose_name_plural = _('Classifiers')
        ordering = ['name']

    slug = models.SlugField(
        allow_unicode=True,
        unique=True,
        verbose_name=_('Slug'),
    )
    name = models.CharField(
        max_length=255,
        verbose_name=_('Name'),
    )

    panels = [
        FieldPanel('name'),
        InlinePanel('terms', label=_('Classifier Terms'))
    ]

    def save(self, *args, **kwargs):
        if not self.slug:
            # Make a slug and suffix a number if it already exists to ensure uniqueness
            newslug = slugify(self.name, allow_unicode=True)
            tmpslug = newslug
            suffix = 1
            while True:
                if not Classifier.objects.filter(slug=tmpslug).exists():
                    self.slug = tmpslug
                    break
                tmpslug = newslug + "-" + str(suffix)
                suffix += 1
        return super().save(*args, **kwargs)

    def __str__(self):
        return self.name


class ClassifierTerm(Orderable, models.Model):
    """
    Term used to categorize a page.
    """
    class Meta:
        verbose_name = _('Classifier Term')
        verbose_name_plural = _('Classifier Terms')

    classifier = ParentalKey(
        Classifier,
        related_name='terms',
        verbose_name=_('Classifier'),
    )
    slug = models.SlugField(
        allow_unicode=True,
        unique=True,
        verbose_name=_('Slug'),
    )
    name = models.CharField(
        max_length=255,
        verbose_name=_('Name'),
    )

    panels = [
        FieldPanel('name'),
    ]

    def save(self, *args, **kwargs):
        if not self.slug:
            # Make a slug and suffix a number if it already exists to ensure uniqueness
            newslug = slugify(self.name, allow_unicode=True)
            tmpslug = newslug
            suffix = 1
            while True:
                if not ClassifierTerm.objects.filter(slug=tmpslug).exists():
                    self.slug = tmpslug
                    break
                tmpslug = newslug + "-" + str(suffix)
                suffix += 1
        return super().save(*args, **kwargs)

    def __str__(self):
        return "{0} > {1}".format(self.classifier.name, self.name)


@register_snippet
class Navbar(models.Model):
    """
    Snippet for site navigation bars (header, main menu, etc.)
    """
    class Meta:
        verbose_name = _('Navigation Bar')

    name = models.CharField(
        max_length=255,
        verbose_name=_('Name'),
    )
    custom_css_class = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Custom CSS Class'),
    )
    custom_id = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Custom ID'),
    )
    menu_items = StreamField(
        NAVIGATION_STREAMBLOCKS,
        verbose_name=_('Navigation links'),
    )

    panels = [
        FieldPanel('name'),
        MultiFieldPanel(
            [
                FieldPanel('custom_css_class'),
                FieldPanel('custom_id'),
            ],
            heading=_('Attributes')
        ),
        StreamFieldPanel('menu_items')
    ]

    def __str__(self):
        return self.name


@register_snippet
class Footer(models.Model):
    """
    Snippet for website footer content.
    """
    class Meta:
        verbose_name = _('Footer')

    name = models.CharField(
        max_length=255,
        verbose_name=_('Name'),
    )
    custom_css_class = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Custom CSS Class'),
    )
    custom_id = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Custom ID'),
    )
    content = StreamField(
        LAYOUT_STREAMBLOCKS,
        verbose_name=_('Content'),
    )

    panels = [
        FieldPanel('name'),
        MultiFieldPanel(
            [
                FieldPanel('custom_css_class'),
                FieldPanel('custom_id'),
            ],
            heading=_('Attributes')
        ),
        StreamFieldPanel('content')
    ]

    def __str__(self):
        return self.name


@register_snippet
class ReusableContent(models.Model):
    """
    Snippet for resusable content in streamfields.
    """
    class Meta:
        verbose_name = _('Reusable Content')
        verbose_name_plural = _('Reusable Content')

    name = models.CharField(
        max_length=255,
        verbose_name=_('Name'),
    )
    content = StreamField(
        LAYOUT_STREAMBLOCKS,
        verbose_name=_('content')
    )

    panels = [
        FieldPanel('name'),
        StreamFieldPanel('content')
    ]

    def __str__(self):
        return self.name


@register_snippet
class ContentWall(models.Model):
    """
    Snippet that restricts access to a page with a modal.
    """
    class Meta:
        verbose_name = _('Content Wall')

    name = models.CharField(
        max_length=255,
        verbose_name=_('Name'),
    )
    content = StreamField(
        LAYOUT_STREAMBLOCKS,
        verbose_name=_('Content'),
    )
    is_dismissible = models.BooleanField(
        default=True,
        verbose_name=_('Dismissible'),
    )
    show_once = models.BooleanField(
        default=True,
        verbose_name=_('Show once'),
        help_text=_('Do not show the content wall to the same user again after it has been closed.')
    )

    panels = [
        MultiFieldPanel(
            [
                FieldPanel('name'),
                FieldPanel('is_dismissible'),
                FieldPanel('show_once'),
            ],
            heading=_('Content Wall')
        ),
        StreamFieldPanel('content'),
    ]

    def __str__(self):
        return self.name


class CoderedEmail(ClusterableModel):
    """
    General purpose abstract clusterable model used for holding email information.
    Most likely this should be subclassed with addition of a ParentalKey.
    """
    class Meta:
        abstract = True
        verbose_name = _('CodeRed Email')

    to_address = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('To Addresses'),
        help_text=_('Separate multiple email addresses with commas.')
    )
    from_address = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('From Address'),
        help_text=_('For example: "sender@example.com" or "Sender Name <sender@example.com>" (without quotes).')  # noqa
    )
    reply_address = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Reply-To Address'),
        help_text=_('Separate multiple email addresses with commas.')
    )
    cc_address = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('CC'),
        help_text=_('Separate multiple email addresses with commas.')
    )
    bcc_address = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('BCC'),
        help_text=_('Separate multiple email addresses with commas.')
    )
    subject = models.CharField(max_length=255, blank=True, verbose_name=_('Subject'))
    body = models.TextField(blank=True, verbose_name=_('Body'))

    panels = (
        [
            MultiFieldPanel(
                [
                    FieldPanel('to_address'),
                    FieldPanel('from_address'),
                    FieldPanel('cc_address'),
                    FieldPanel('bcc_address'),
                    FieldPanel('subject'),
                    FieldPanel('body'),
                ],
                _('Email Message')
            ),
        ])

    def __str__(self):
        return self.subject



"""
Bases, mixins, and utilities for blocks.
"""

from django import forms
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from wagtail.core import blocks
from wagtail.core.models import Collection
from wagtail.core.utils import resolve_model_string
from wagtail.documents.blocks import DocumentChooserBlock

from coderedcms.settings import cr_settings


class ClassifierTermChooserBlock(blocks.FieldBlock):
    """
    Enables choosing a ClassifierTerm in the streamfield.
    Lazy loads the target_model from the string to avoid recursive imports.
    """
    widget = forms.Select

    def __init__(self, required=False, label=None, help_text=None, *args, **kwargs):
        self._required = required
        self._help_text = help_text
        self._label = label
        super().__init__(*args, **kwargs)

    @cached_property
    def target_model(self):
        return resolve_model_string('coderedcms.ClassifierTerm')

    @cached_property
    def field(self):
        return forms.ModelChoiceField(
            queryset=self.target_model.objects.all().order_by('classifier__name', 'name'),
            widget=self.widget,
            required=self._required,
            label=self._label,
            help_text=self._help_text,
        )

    def to_python(self, value):
        """
        Convert the serialized value back into a python object.
        """
        if isinstance(value, int):
            return self.target_model.objects.get(pk=value)
        return value

    def get_prep_value(self, value):
        """
        Serialize the model in a form suitable for wagtail's JSON-ish streamfield
        """
        if isinstance(value, self.target_model):
            return value.pk
        return value


class CollectionChooserBlock(blocks.FieldBlock):
    """
    Enables choosing a wagtail Collection in the streamfield.
    """
    target_model = Collection
    widget = forms.Select

    def __init__(self, required=False, label=None, help_text=None, *args, **kwargs):
        self._required = required
        self._help_text = help_text
        self._label = label
        super().__init__(*args, **kwargs)

    @cached_property
    def field(self):
        return forms.ModelChoiceField(
            queryset=self.target_model.objects.all().order_by('name'),
            widget=self.widget,
            required=self._required,
            label=self._label,
            help_text=self._help_text,
        )

    def to_python(self, value):
        """
        Convert the serialized value back into a python object.
        """
        if isinstance(value, int):
            return self.target_model.objects.get(pk=value)
        return value

    def get_prep_value(self, value):
        """
        Serialize the model in a form suitable for wagtail's JSON-ish streamfield
        """
        if isinstance(value, self.target_model):
            return value.pk
        return value


class ButtonMixin(blocks.StructBlock):
    """
    Standard style and size options for buttons.
    """
    button_title = blocks.CharBlock(
        max_length=255,
        required=True,
        label=_('Button Title'),
    )
    button_style = blocks.ChoiceBlock(
        choices=cr_settings['FRONTEND_BTN_STYLE_CHOICES'],
        default=cr_settings['FRONTEND_BTN_STYLE_DEFAULT'],
        required=False,
        label=_('Button Style'),
    )
    button_size = blocks.ChoiceBlock(
        choices=cr_settings['FRONTEND_BTN_SIZE_CHOICES'],
        default=cr_settings['FRONTEND_BTN_SIZE_DEFAULT'],
        required=False,
        label=_('Button Size'),
    )


class CoderedAdvSettings(blocks.StructBlock):
    """
    Common fields each block should have,
    which are hidden under the block's "Advanced Settings" dropdown.
    """
    # placeholder, real value get set in __init__()
    custom_template = blocks.Block()

    custom_css_class = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Custom CSS Class'),
    )
    custom_id = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Custom ID'),
    )

    class Meta:
        form_template = 'wagtailadmin/block_forms/base_block_settings_struct.html'
        label = _('Advanced Settings')

    def __init__(self, local_blocks=None, template_choices=None, **kwargs):
        if not local_blocks:
            local_blocks = ()

        local_blocks += (
            (
                'custom_template',
                blocks.ChoiceBlock(
                    choices=template_choices,
                    default=None,
                    required=False,
                    label=_('Template'))
            ),
        )

        super().__init__(local_blocks, **kwargs)


class CoderedAdvTrackingSettings(CoderedAdvSettings):
    """
    CoderedAdvSettings plus additional tracking fields.
    """
    ga_tracking_event_category = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Tracking Event Category'),
    )
    ga_tracking_event_label = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Tracking Event Label'),
    )


class CoderedAdvColumnSettings(CoderedAdvSettings):
    """
    BaseBlockSettings plus additional column fields.
    """
    column_breakpoint = blocks.ChoiceBlock(
        choices=cr_settings['FRONTEND_COL_BREAK_CHOICES'],
        default=cr_settings['FRONTEND_COL_BREAK_DEFAULT'],
        required=False,
        verbose_name=_('Column Breakpoint'),
        help_text=_('Screen size at which the column will expand horizontally or stack vertically.'),  # noqa
    )


class BaseBlock(blocks.StructBlock):
    """
    Common attributes for all blocks used in CodeRed CMS.
    """
    # subclasses can override this to determine the advanced settings class
    advsettings_class = CoderedAdvSettings

    # placeholder, real value get set in __init__() from advsettings_class
    settings = blocks.Block()

    def __init__(self, local_blocks=None, **kwargs):
        """
        Construct and inject settings block, then initialize normally.
        """
        klassname = self.__class__.__name__.lower()
        choices = cr_settings['FRONTEND_TEMPLATES_BLOCKS'].get('*', ()) + \
            cr_settings['FRONTEND_TEMPLATES_BLOCKS'].get(klassname, ())

        if not local_blocks:
            local_blocks = ()

        local_blocks += (('settings', self.advsettings_class(template_choices=choices)),)

        super().__init__(local_blocks, **kwargs)

    def render(self, value, context=None):
        template = value['settings']['custom_template']

        if not template:
            template = self.get_template(context=context)
            if not template:
                return self.render_basic(value, context=context)

        if context is None:
            new_context = self.get_context(value)
        else:
            new_context = self.get_context(value, parent_context=dict(context))

        return mark_safe(render_to_string(template, new_context))


class BaseLayoutBlock(BaseBlock):
    """
    Common attributes for all blocks used in CodeRed CMS.
    """
    # Subclasses can override this to provide a default list of blocks for the content.
    content_streamblocks = []

    def __init__(self, local_blocks=None, **kwargs):
        if not local_blocks and self.content_streamblocks:
            local_blocks = self.content_streamblocks

        if local_blocks:
            local_blocks = (('content', blocks.StreamBlock(local_blocks, label=_('Content'))),)

        super().__init__(local_blocks, **kwargs)


class LinkStructValue(blocks.StructValue):
    """
    Generates a URL for blocks with multiple link choices.
    """
    @property
    def url(self):
        page = self.get('page_link')
        doc = self.get('doc_link')
        ext = self.get('other_link')
        if page and ext:
            return "{0}{1}".format(page.url, ext)
        elif page:
            return page.url
        elif doc:
            return doc.url
        else:
            return ext


class BaseLinkBlock(BaseBlock):
    """
    Common attributes for creating a link within the CMS.
    """
    page_link = blocks.PageChooserBlock(
        required=False,
        label=_('Page link'),
    )
    doc_link = DocumentChooserBlock(
        required=False,
        label=_('Document link'),
    )
    other_link = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_('Other link'),
    )

    advsettings_class = CoderedAdvTrackingSettings

    class Meta:
        value_class = LinkStructValue



"""
Creatable pages used in CodeRed CMS.
"""
from modelcluster.fields import ParentalKey
from coderedcms.forms import CoderedFormField
from coderedcms.models import (
    CoderedArticlePage,
    CoderedArticleIndexPage,
    CoderedEmail,
    CoderedFormPage,
    CoderedWebPage
)


class ArticlePage(CoderedArticlePage):
    """
    Article, suitable for news or blog content.
    """
    class Meta:
        verbose_name = 'Article'
        ordering = ['-first_published_at']

    # Only allow this page to be created beneath an ArticleIndexPage.
    parent_page_types = ['website.ArticleIndexPage']

    template = 'coderedcms/pages/article_page.html'
    search_template = 'coderedcms/pages/article_page.search.html'


class ArticleIndexPage(CoderedArticleIndexPage):
    """
    Shows a list of article sub-pages.
    """
    class Meta:
        verbose_name = 'Article Landing Page'

    # Override to specify custom index ordering choice/default.
    index_query_pagemodel = 'website.ArticlePage'

    # Only allow ArticlePages beneath this page.
    subpage_types = ['website.ArticlePage']

    template = 'coderedcms/pages/article_index_page.html'


class FormPage(CoderedFormPage):
    """
    A page with an html <form>.
    """
    class Meta:
        verbose_name = 'Form'

    template = 'coderedcms/pages/form_page.html'


class FormPageField(CoderedFormField):
    """
    A field that links to a FormPage.
    """
    class Meta:
        ordering = ['sort_order']

    page = ParentalKey('FormPage', related_name='form_fields')


class FormConfirmEmail(CoderedEmail):
    """
    Sends a confirmation email after submitting a FormPage.
    """
    page = ParentalKey('FormPage', related_name='confirmation_emails')


class WebPage(CoderedWebPage):
    """
    General use page with featureful streamfield and SEO attributes.
    """
    class Meta:
        verbose_name = 'Web Page'

    template = 'coderedcms/pages/web_page.html'



"""
Base and abstract pages used in CodeRed CMS.
"""

import json
import logging
import os
import warnings
from datetime import datetime
from typing import Optional, TYPE_CHECKING

import geocoder
from django import forms
from django.conf import settings
from django.contrib import messages
from django.core.files.uploadedfile import InMemoryUploadedFile, TemporaryUploadedFile
from django.core.files.storage import FileSystemStorage
from django.core.mail import EmailMessage
from django.core.paginator import Paginator, InvalidPage, EmptyPage, PageNotAnInteger
from django.core.serializers.json import DjangoJSONEncoder
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.http import JsonResponse, HttpResponseRedirect
from django.shortcuts import render, redirect
from django.template import Context, Template
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.html import strip_tags
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from eventtools.models import BaseEvent, BaseOccurrence
from icalendar import Event as ICalEvent
from modelcluster.fields import ParentalKey, ParentalManyToManyField
from modelcluster.tags import ClusterTaggableManager
from pathlib import Path
from taggit.models import TaggedItemBase
from wagtail.admin.edit_handlers import (
    FieldPanel,
    FieldRowPanel,
    InlinePanel,
    MultiFieldPanel,
    ObjectList,
    PageChooserPanel,
    StreamFieldPanel,
    TabbedInterface
)
from wagtail.core import hooks
from wagtail.core.fields import StreamField
from wagtail.core.models import Orderable, PageBase, Page, Site
from wagtail.core.utils import resolve_model_string
from wagtail.contrib.forms.edit_handlers import FormSubmissionsPanel
from wagtail.contrib.forms.forms import WagtailAdminFormPageForm
from wagtail.images import get_image_model_string
from wagtail.images.edit_handlers import ImageChooserPanel
from wagtail.contrib.forms.models import FormSubmission
from wagtail.search import index
from wagtail.utils.decorators import cached_classmethod
from wagtailcache.cache import WagtailCacheMixin
from wagtailseo.models import SeoMixin, TwitterCard
from wagtailseo.utils import get_struct_data_images, StructDataEncoder

from coderedcms import utils
from coderedcms.blocks import (
    CONTENT_STREAMBLOCKS,
    LAYOUT_STREAMBLOCKS,
    STREAMFORM_BLOCKS,
    ContentWallBlock,
)
from coderedcms.fields import ColorField
from coderedcms.forms import CoderedFormBuilder, CoderedSubmissionsListView
from coderedcms.models.snippet_models import ClassifierTerm
from coderedcms.models.wagtailsettings_models import (
    GeneralSettings,
    GoogleApiSettings,
    LayoutSettings,
)
from coderedcms.wagtail_flexible_forms.blocks import FormFieldBlock, FormStepBlock
from coderedcms.wagtail_flexible_forms.models import (
    Step,
    Steps,
    StreamFormMixin,
    StreamFormJSONEncoder,
    SessionFormSubmission,
    SubmissionRevision,
)
from coderedcms.settings import cr_settings
from coderedcms.widgets import ClassifierSelectWidget


if TYPE_CHECKING:
    from wagtail.images.models import AbstractImage


logger = logging.getLogger('coderedcms')


CODERED_PAGE_MODELS = []


def get_page_models():
    return CODERED_PAGE_MODELS


class CoderedPageMeta(PageBase):
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        if 'search_db_include' not in dct:
            cls.search_db_include = False
        if 'search_db_boost' not in dct:
            cls.search_db_boost = 0
        if 'search_filterable' not in dct:
            cls.search_filterable = False
        if 'search_name' not in dct:
            cls.search_name = cls._meta.verbose_name
        if 'search_name_plural' not in dct:
            cls.search_name_plural = cls._meta.verbose_name_plural
        if 'search_template' not in dct:
            cls.search_template = 'coderedcms/pages/search_result.html'
        if not cls._meta.abstract:
            CODERED_PAGE_MODELS.append(cls)


class CoderedTag(TaggedItemBase):
    class Meta:
        verbose_name = _('CodeRed Tag')
    content_object = ParentalKey('coderedcms.CoderedPage', related_name='tagged_items')


class CoderedPage(WagtailCacheMixin, SeoMixin, Page, metaclass=CoderedPageMeta):
    """
    General use page with caching, templating, and SEO functionality.
    All pages should inherit from this.
    """
    class Meta:
        verbose_name = _('CodeRed Page')

    # Do not allow this page type to be created in wagtail admin
    is_creatable = False

    # Templates
    # The page will render the following templates under certain conditions:
    #
    # template = ''
    # ajax_template = ''
    # search_template = ''

    ###############
    # Content fields
    ###############

    cover_image = models.ForeignKey(
        get_image_model_string(),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name='+',
        verbose_name=_('Cover image'),
    )

    ###############
    # Index fields
    ###############

    # Subclasses can override this to enabled index features by default.
    index_show_subpages_default = False

    # Subclasses can override this to query on a specific
    # page model, rather than the default wagtail Page.
    index_query_pagemodel = 'coderedcms.CoderedPage'

    # Subclasses can override these fields to enable custom
    # ordering based on specific subpage fields.
    index_order_by_default = ''
    index_order_by_choices = (
        ('', _('Default Ordering')),
        ('-first_published_at', _('Date first published, newest to oldest')),
        ('first_published_at', _('Date first published, oldest to newest')),
        ('-last_published_at', _('Date updated, newest to oldest')),
        ('last_published_at', _('Date updated, oldest to newest')),
        ('title', _('Title, alphabetical')),
        ('-title', _('Title, reverse alphabetical')),
    )
    index_show_subpages = models.BooleanField(
        default=index_show_subpages_default,
        verbose_name=_('Show list of child pages')
    )
    index_order_by = models.CharField(
        max_length=255,
        choices=index_order_by_choices,
        default=index_order_by_default,
        blank=True,
        verbose_name=_('Order child pages by'),
    )
    index_num_per_page = models.PositiveIntegerField(
        default=10,
        verbose_name=_('Number per page'),
    )
    index_classifiers = ParentalManyToManyField(
        'coderedcms.Classifier',
        blank=True,
        verbose_name=_('Filter child pages by'),
        help_text=_('Enable filtering child pages by these classifiers.'),
    )

    ###############
    # Layout fields
    ###############

    custom_template = models.CharField(
        blank=True,
        max_length=255,
        choices=None,
        verbose_name=_('Template')
    )

    ###############
    # SEO overrides
    ###############

    seo_image_sources = [
        "og_image",
        "cover_image",
    ]

    ###############
    # Classify
    ###############

    classifier_terms = ParentalManyToManyField(
        'coderedcms.ClassifierTerm',
        blank=True,
        verbose_name=_('Classifiers'),
        help_text=_('Categorize and group pages together with classifiers. Used to organize and filter pages across the site.'),  # noqa
    )
    tags = ClusterTaggableManager(
        through=CoderedTag,
        blank=True,
        verbose_name=_('Tags'),
        help_text=_('Used to organize pages across the site.'),
    )

    ###############
    # Settings
    ###############

    content_walls = StreamField(
        [
            ('content_wall', ContentWallBlock())
        ],
        blank=True,
        verbose_name=_('Content Walls')
    )

    ###############
    # Search
    ###############

    search_fields = [
        index.SearchField('title', partial_match=True, boost=3),
        index.SearchField('seo_title', partial_match=True, boost=3),
        index.SearchField('search_description', boost=2),
        index.FilterField('title'),
        index.FilterField('id'),
        index.FilterField('live'),
        index.FilterField('owner'),
        index.FilterField('content_type'),
        index.FilterField('path'),
        index.FilterField('depth'),
        index.FilterField('locked'),
        index.FilterField('first_published_at'),
        index.FilterField('last_published_at'),
        index.FilterField('latest_revision_created_at'),
        index.FilterField('index_show_subpages'),
        index.FilterField('index_order_by'),
        index.FilterField('custom_template'),
        index.FilterField('classifier_terms'),
    ]

    ###############
    # Panels
    ###############

    content_panels = Page.content_panels + [
        ImageChooserPanel('cover_image'),
    ]

    body_content_panels = []

    bottom_content_panels = []

    classify_panels = [
        FieldPanel('classifier_terms', widget=ClassifierSelectWidget()),
        FieldPanel('tags'),
    ]

    layout_panels = [
        MultiFieldPanel(
            [
                FieldPanel('custom_template')
            ],
            heading=_('Visual Design')
        ),
        MultiFieldPanel(
            [
                FieldPanel('index_show_subpages'),
                FieldPanel('index_num_per_page'),
                FieldPanel('index_order_by'),
                FieldPanel('index_classifiers', widget=forms.CheckboxSelectMultiple()),
            ],
            heading=_('Show Child Pages')
        )
    ]

    promote_panels = SeoMixin.seo_meta_panels + SeoMixin.seo_struct_panels

    settings_panels = Page.settings_panels + [
        StreamFieldPanel('content_walls'),
    ]

    integration_panels = []

    def __init__(self, *args, **kwargs):
        """
        Inject custom choices and defaults into the form fields
        to enable customization by subclasses.
        """
        super().__init__(*args, **kwargs)
        klassname = self.__class__.__name__.lower()
        template_choices = cr_settings['FRONTEND_TEMPLATES_PAGES'].get('*', ()) + \
            cr_settings['FRONTEND_TEMPLATES_PAGES'].get(klassname, ())

        self._meta.get_field('index_order_by').choices = self.index_order_by_choices
        self._meta.get_field('custom_template').choices = template_choices
        if not self.id:
            self.index_order_by = self.index_order_by_default
            self.index_show_subpages = self.index_show_subpages_default

    @cached_classmethod
    def get_edit_handler(cls):
        """
        Override to "lazy load" the panels overridden by subclasses.
        """
        panels = [
            ObjectList(
                cls.content_panels + cls.body_content_panels + cls.bottom_content_panels,
                heading=_('Content')
            ),
            ObjectList(cls.classify_panels, heading=_('Classify')),
            ObjectList(cls.layout_panels, heading=_('Layout')),
            ObjectList(cls.promote_panels, heading=_('SEO'), classname="seo"),
            ObjectList(cls.settings_panels, heading=_('Settings'), classname="settings"),
        ]

        if cls.integration_panels:
            panels.append(ObjectList(
                cls.integration_panels,
                heading='Integrations',
                classname='integrations'
            ))

        return TabbedInterface(panels).bind_to(model=cls)

    @property
    def seo_logo(self) -> "Optional[AbstractImage]":
        """
        Override method in SeoMixin.
        Gets the primary logo of the organization.
        """
        logo = super().seo_logo
        if logo:
            return logo
        else:
            layout_settings = LayoutSettings.for_site(self.get_site())
            if layout_settings.logo:
                return layout_settings.logo
        return None

    @property
    def seo_image(self) -> "Optional[AbstractImage]":
        """
        Override method in SeoMixin.
        Fallback to logo if opengraph image is not specified.
        """
        img = super().seo_image
        if img is None:
            return self.seo_logo
        return img

    @property
    def seo_twitter_card_content(self) -> str:
        """
        Override of method in SeoMixin.
        Show a large twitter card if the page has an image set.
        """
        if self.seo_image:
            return TwitterCard.LARGE.value
        return self.seo_twitter_card.value

    def get_template(self, request, *args, **kwargs):
        """
        Override parent to serve different templates based on querystring.
        """
        if self.custom_template:
            return self.custom_template

        return super().get_template(request, args, kwargs)

    def get_index_children(self):
        """
        Returns query of subpages as defined by `index_` variables.
        """
        if self.index_query_pagemodel:
            querymodel = resolve_model_string(self.index_query_pagemodel, self._meta.app_label)
            query = querymodel.objects.child_of(self).live()
        else:
            query = self.get_children().live()
        if self.index_order_by:
            return query.order_by(self.index_order_by)
        return query

    def get_content_walls(self, check_child_setting=True):
        current_content_walls = []
        if check_child_setting:
            for wall in self.content_walls:
                if wall.value['show_content_wall_on_children']:
                    current_content_walls.append(wall.value)
        else:
            current_content_walls = self.content_walls

        try:
            return list(current_content_walls) + self.get_parent().specific.get_content_walls()
        except AttributeError:
            return list(current_content_walls)

    def get_context(self, request, *args, **kwargs):
        """
        Add child pages and paginated child pages to context.
        """
        context = super().get_context(request)

        if self.index_show_subpages:
            # Get child pages
            all_children = self.get_index_children()
            # Filter by classifier terms if applicable
            if len(request.GET) > 0 and self.index_classifiers.exists():
                # Look up comma separated ClassifierTerm slugs i.e. `/?c=term1-slug,term2-slug`
                terms = []
                get_c = request.GET.get('c', None)
                if get_c:
                    terms = get_c.split(',')
                # Else look up individual querystrings i.e. `/?classifier-slug=term1-slug`
                else:
                    for classifier in self.index_classifiers.all().only('slug'):
                        get_term = request.GET.get(classifier.slug, None)
                        if get_term:
                            terms.append(get_term)
                if len(terms) > 0:
                    selected_terms = ClassifierTerm.objects.filter(slug__in=terms)
                    context['selected_terms'] = selected_terms
                    if len(selected_terms) > 0:
                        try:
                            for term in selected_terms:
                                all_children = all_children.filter(classifier_terms=term)
                        except AttributeError:
                            logger.warning(
                                "Tried to filter by ClassifierTerm, but <%s.%s ('%s')>.get_index_children() did not return a queryset or is not a queryset of CoderedPage models.",  # noqa
                                self._meta.app_label,
                                self.__class__.__name__,
                                self.title
                            )
            paginator = Paginator(all_children, self.index_num_per_page)
            pagenum = request.GET.get('p', 1)
            try:
                paged_children = paginator.page(pagenum)
            except (PageNotAnInteger, EmptyPage, InvalidPage) as e:  # noqa
                paged_children = paginator.page(1)

            context['index_paginated'] = paged_children
            context['index_children'] = all_children
        context['content_walls'] = self.get_content_walls(check_child_setting=False)
        return context


###############################################################################
# Abstract pages providing pre-built common website functionality, suitable for subclassing.
# These are abstract so subclasses can override fields if desired.
###############################################################################


class CoderedWebPage(CoderedPage):
    """
    Provides a body and body-related functionality.
    This is abstract so that subclasses can override the body StreamField.
    """
    class Meta:
        verbose_name = _('CodeRed Web Page')
        abstract = True

    template = 'coderedcms/pages/web_page.html'

    # Child pages should override based on what blocks they want in the body.
    # Default is LAYOUT_STREAMBLOCKS which is the fullest editor experience.
    body = StreamField(LAYOUT_STREAMBLOCKS, null=True, blank=True)

    # Search fields
    search_fields = (
        CoderedPage.search_fields +
        [index.SearchField('body')]
    )

    # Panels
    body_content_panels = [
        StreamFieldPanel('body'),
    ]

    @property
    def body_preview(self):
        """
        A shortened version of the body without HTML tags.
        """
        # add spaces between tags for legibility
        body = str(self.body).replace('>', '> ')
        # strip tags
        body = strip_tags(body)
        # truncate and add ellipses
        preview = body[:200] + "..." if len(body) > 200 else body
        return mark_safe(preview)

    @property
    def page_ptr(self):
        """
        Overwrite of `page_ptr` to make it compatible with wagtailimportexport.
        """
        return self.base_page_ptr

    @page_ptr.setter
    def page_ptr(self, value):
        self.base_page_ptr = value


class CoderedArticlePage(CoderedWebPage):
    """
    Article, suitable for news or blog content.
    """
    class Meta:
        verbose_name = _('CodeRed Article')
        abstract = True

    template = 'coderedcms/pages/article_page.html'

    # Override body to provide simpler content
    body = StreamField(CONTENT_STREAMBLOCKS, null=True, blank=True)

    caption = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Caption'),
    )
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        editable=True,
        on_delete=models.SET_NULL,
        verbose_name=_('Author'),
    )
    author_display = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Display author as'),
        help_text=_('Override how the author’s name displays on this article.'),
    )
    date_display = models.DateField(
        null=True,
        blank=True,
        verbose_name=_('Display publish date'),
    )

    def get_author_name(self):
        warnings.warn(
            ("CoderedArticlePage.get_author_name has been replaced with "
             "CoderedArticlePage.seo_author"),
            DeprecationWarning,
        )
        return self.seo_author

    @property
    def seo_author(self) -> str:
        """
        Override of method in SeoMixin.
        Gets author name using a fallback.
        """
        if self.author_display:
            return self.author_display
        if self.author:
            return self.author.get_full_name()
        if self.owner:
            return self.owner.get_full_name()
        return ""

    def get_pub_date(self):
        warnings.warn(
            ("CoderedArticlePage.get_pub_date has been replaced with "
             "CoderedArticlePage.seo_published_at"),
            DeprecationWarning,
        )
        return self.seo_published_at

    @property
    def seo_published_at(self) -> datetime:
        """
        Override of method in SeoMixin.
        Gets published date.
        """
        if self.date_display:
            return self.date_display
        return self.first_published_at

    def get_description(self):
        warnings.warn(
            ("CoderedArticlePage.get_description has been replaced with "
             "CoderedArticlePage.seo_description"),
            DeprecationWarning,
        )
        return self.seo_description

    @property
    def seo_description(self) -> str:
        """
        Override of method in SeoMixin.
        Gets the description using a fallback.
        """
        if self.search_description:
            return self.search_description
        if self.caption:
            return self.caption
        if self.body_preview:
            return self.body_preview
        return ""

    search_fields = (
        CoderedWebPage.search_fields +
        [
            index.SearchField('caption', boost=2),
            index.FilterField('author'),
            index.FilterField('author_display'),
            index.FilterField('date_display'),
        ]
    )

    content_panels = CoderedWebPage.content_panels + [
        FieldPanel('caption'),
        MultiFieldPanel(
            [
                FieldPanel('author'),
                FieldPanel('author_display'),
                FieldPanel('date_display'),
            ],
            _('Publication Info')
        )
    ]


class CoderedArticleIndexPage(CoderedWebPage):
    """
    Shows a list of article sub-pages.
    """
    class Meta:
        verbose_name = _('CodeRed Article Index Page')
        abstract = True

    template = 'coderedcms/pages/article_index_page.html'

    index_show_subpages_default = True

    index_order_by_default = '-date_display'
    index_order_by_choices = (('-date_display', 'Display publish date, newest first'),) + \
        CoderedWebPage.index_order_by_choices

    show_images = models.BooleanField(
        default=True,
        verbose_name=_('Show images'),
    )
    show_captions = models.BooleanField(
        default=True,
    )
    show_meta = models.BooleanField(
        default=True,
        verbose_name=_('Show author and date info'),
    )
    show_preview_text = models.BooleanField(
        default=True,
        verbose_name=_('Show preview text'),
    )

    layout_panels = CoderedWebPage.layout_panels + [
        MultiFieldPanel(
            [
                FieldPanel('show_images'),
                FieldPanel('show_captions'),
                FieldPanel('show_meta'),
                FieldPanel('show_preview_text'),
            ],
            heading=_('Child page display')
        ),
    ]


class CoderedEventPage(CoderedWebPage, BaseEvent):
    class Meta:
        verbose_name = _('CodeRed Event')
        abstract = True

    calendar_color = ColorField(
        blank=True,
        help_text=_('The color that the event will use when displayed on a calendar.'),
    )
    address = models.TextField(
        blank=True,
        verbose_name=_("Address")
    )
    content_panels = CoderedWebPage.content_panels + [
        MultiFieldPanel(
            [
                FieldPanel('calendar_color'),
                FieldPanel('address'),
            ],
            heading=_('Event information')
        ),
        InlinePanel(
            'occurrences',
            min_num=1,
            heading=_("Dates and times"),
        ),
    ]

    @property
    def upcoming_occurrences(self):
        """
        Returns the next x occurrences for this event.
        By default, it returns 10.
        """
        return self.query_occurrences(num_of_instances_to_return=10)

    @property
    def most_recent_occurrence(self):
        """
        Gets the next upcoming, or last occurrence if the event has no more occurrences.
        """

        try:
            noc = self.next_occurrence()
            if noc:
                return noc
            aoc = []
            for occurrence in self.occurrences.all():
                aoc += [instance for instance in occurrence.all_occurrences()]
            if len(aoc) > 0:
                return aoc[-1]  # last one in the list

        except AttributeError:
            # Triggers when a preview is initiated on an
            # EventPage because it uses a FakeQuerySet object.
            # Here we manually compute the next_occurrence
            occurrences = [e.next_occurrence() for e in self.occurrences.all()]
            if occurrences:
                return sorted(occurrences, key=lambda tup: tup[0])[0]

    @property
    def seo_struct_event_dict(self) -> dict:
        next_occ = self.most_recent_occurrence
        sd_dict = {
            "@context": "https://schema.org/",
            "@type": "Event",
            "name": self.title,
            "description": self.seo_description,
            "startDate": next_occ.start,
            "endDate": next_occ.end,
            "mainEntityOfPage": {
                "@type": "WebPage",
                "@id": self.get_full_url,
            },
        }

        if self.seo_image:
            sd_dict.update({"image": get_struct_data_images(self.seo_image)})

        if self.address:
            sd_dict.update({
                "location": {
                    "@type": "Place",
                    "name": self.title,
                    "address": {
                        "@type": "PostalAddress",
                        "streetAddress": self.address,
                    },
                },
            })

        return sd_dict

    @property
    def seo_struct_event_json(self) -> str:
        return json.dumps(self.seo_struct_event_dict, cls=StructDataEncoder)

    def query_occurrences(self, num_of_instances_to_return=None, **kwargs):
        """
        Returns a list of all upcoming event instances for the specified query.
        For more information on what you can query with, visit
        https://github.com/gregplaysguitar/django-eventtools
        """
        event_instances = []
        occurrence_kwargs = {
            'from_date': kwargs.get('from_date', timezone.now().date())
        }

        if 'limit' in kwargs:
            if kwargs['limit'] is not None:
                # Limit the number of event instances that will be
                # generated per occurrence rule to 10, if not otherwise specified.
                occurrence_kwargs['limit'] = kwargs.get('limit', 10)

        # For each occurrence rule in all of the occurrence rules for this event.
        for occurrence in self.occurrences.all():

            # Add the qualifying generated event instances to the list.
            event_instances += [
                instance for instance in occurrence.all_occurrences(**occurrence_kwargs)]

        # Sort all the events by the date that they start
        event_instances.sort(key=lambda d: d[0])

        # Return the event instances, possibly spliced if num_instances_to_return is set.
        return event_instances[:num_of_instances_to_return] if num_of_instances_to_return else event_instances  # noqa

    def convert_to_ical_format(self, dt_start=None, dt_end=None, occurrence=None):
        ical_event = ICalEvent()
        ical_event.add('summary', self.title)
        if self.address:
            ical_event.add('location', self.address)

        if dt_start:
            ical_event.add('dtstart', dt_start)

            if dt_end:
                ical_event.add('dtend', dt_end)

        if occurrence:
            freq = occurrence.repeat.split(":")[1] if occurrence.repeat else None
            repeat_until = occurrence.repeat_until.strftime(
                "%Y%m%dT000000Z") if occurrence.repeat_until else None

            ical_event.add('dtstart', occurrence.start)

            if occurrence.end:
                ical_event.add('dtend', occurrence.end)

            if freq:
                ical_event.add('RRULE', freq, encode=False)

            if repeat_until:
                ical_event.add('until', repeat_until)

        return ical_event

    def create_single_ical(self, dt_start, dt_end=None):
        return self.convert_to_ical_format(dt_start=dt_start, dt_end=dt_end)

    def create_recurring_ical(self):
        events = []
        for occurrence in self.occurrences.all():
            events.append(self.convert_to_ical_format(occurrence=occurrence))
        return events


class DefaultCalendarViewChoices():
    MONTH = 'month'
    AGENDA_WEEK = 'agendaWeek'
    AGENDA_DAY = 'agendaDay'
    LIST_MONTH = 'listMonth'

    CHOICES = (
        ('', _('No calendar')),
        (MONTH, _('Monthly Calendar')),
        (AGENDA_WEEK, _('Weekly Calendar')),
        (AGENDA_DAY, _('Daily Calendar')),
        (LIST_MONTH, _('Calendar List View')),
    )


class CoderedEventIndexPage(CoderedWebPage):
    """
    Shows a list of event sub-pages.
    """
    class Meta:
        verbose_name = _('CodeRed Event Index Page')
        abstract = True

    template = 'coderedcms/pages/event_index_page.html'

    index_show_subpages_default = True

    index_order_by_default = 'next_occurrence'
    index_order_by_choices = (
        ('next_occurrence', 'Display next occurrence, soonest first'),
    ) + CoderedWebPage.index_order_by_choices

    default_calendar_view = models.CharField(
        blank=True,
        choices=DefaultCalendarViewChoices.CHOICES,
        max_length=255,
        verbose_name=_('Calendar Style'),
        help_text=_('The default look of the calendar on this page.')
    )

    layout_panels = CoderedWebPage.layout_panels + [
        FieldPanel('default_calendar_view'),
    ]

    def get_index_children(self):
        if self.index_query_pagemodel and self.index_order_by == 'next_occurrence':
            querymodel = resolve_model_string(self.index_query_pagemodel, self._meta.app_label)
            qs = querymodel.objects.child_of(self).live()
            # filter out events that don't have a next_occurrence
            upcoming = []
            for event in qs.all():
                if event.next_occurrence():
                    upcoming.append(event)
            # sort the events by next_occurrence
            return sorted(upcoming, key=lambda e: e.next_occurrence())

        return super().get_index_children()

    def get_calendar_events(self, start, end):
        # start with all child events, regardless of get_index_children rules.
        querymodel = resolve_model_string(self.index_query_pagemodel, self._meta.app_label)
        qs = querymodel.objects.child_of(self).live()
        event_instances = []
        for event in qs:
            occurrences = event.query_occurrences(limit=None, from_date=start, to_date=end)
            for occurrence in occurrences:
                event_data = {
                    'title': event.title,
                    'start': occurrence[0].strftime('%Y-%m-%dT%H:%M:%S'),
                    'end': occurrence[1].strftime('%Y-%m-%dT%H:%M:%S') if occurrence[1] else "",
                    'description': "",
                }
                if event.url:
                    event_data['url'] = event.url
                if event.calendar_color:
                    event_data['backgroundColor'] = event.calendar_color
                event_instances.append(event_data)
        return event_instances


class CoderedEventOccurrence(Orderable, BaseOccurrence):
    class Meta:
        verbose_name = _('CodeRed Event Occurrence')
        abstract = True


class CoderedFormMixin(models.Model):
    class Meta:
        abstract = True

    submissions_list_view_class = CoderedSubmissionsListView
    encoder = DjangoJSONEncoder

    # Custom codered fields
    to_address = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Email form submissions to'),
        help_text=_('Optional - email form submissions to this address. Separate multiple addresses by comma.'),  # noqa
    )
    reply_address = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Reply-to address'),
        help_text=_('Optional - to reply to the submitter, specify the email field here. For example, if a form field above is labeled "Your Email", enter: {{ your_email }}'),  # noqa
    )
    subject = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Subject'),
    )
    save_to_database = models.BooleanField(
        default=True,
        verbose_name=_('Save form submissions'),
        help_text=_('Submissions are saved to database and can be exported at any time.')
    )
    thank_you_page = models.ForeignKey(
        'wagtailcore.Page',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name='+',
        verbose_name=_('Thank you page'),
        help_text=_('The page users are redirected to after submitting the form.'),
    )
    button_text = models.CharField(
        max_length=255,
        default=_('Submit'),
        verbose_name=_('Button text'),
    )
    button_style = models.CharField(
        blank=True,
        choices=cr_settings['FRONTEND_BTN_STYLE_CHOICES'],
        default=cr_settings["FRONTEND_BTN_STYLE_DEFAULT"],
        max_length=255,
        verbose_name=_('Button style'),
    )
    button_size = models.CharField(
        blank=True,
        choices=cr_settings['FRONTEND_BTN_SIZE_CHOICES'],
        default=cr_settings["FRONTEND_BTN_SIZE_DEFAULT"],
        max_length=255,
        verbose_name=_('Button Size'),
    )
    button_css_class = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Button CSS class'),
        help_text=_('Custom CSS class applied to the submit button.'),
    )
    form_css_class = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Form CSS Class'),
        help_text=_('Custom CSS class applied to <form> element.'),
    )
    form_id = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('Form ID'),
        help_text=_('Custom ID applied to <form> element.'),
    )
    form_golive_at = models.DateTimeField(
        blank=True,
        null=True,
        verbose_name=_('Form go live date/time'),
        help_text=_('Date and time when the FORM goes live on the page.'),
    )
    form_expire_at = models.DateTimeField(
        blank=True,
        null=True,
        verbose_name=_('Form expiry date/time'),
        help_text=_('Date and time when the FORM will no longer be available on the page.'),
    )
    spam_protection = models.BooleanField(
        default=True,
        verbose_name=_('Spam Protection'),
        help_text=_('When enabled, the CMS will filter out spam form submissions for this page.')
    )

    body_content_panels = [
        MultiFieldPanel(
            [
                PageChooserPanel('thank_you_page'),
                FieldPanel('button_text'),
                FieldPanel('button_style'),
                FieldPanel('button_size'),
                FieldPanel('button_css_class'),
                FieldPanel('form_css_class'),
                FieldPanel('form_id'),
            ],
            _('Form Settings')
        ),
        MultiFieldPanel(
            [
                FieldPanel('save_to_database'),
                FieldPanel('to_address'),
                FieldPanel('reply_address'),
                FieldPanel('subject'),
            ],
            _('Form Submissions')
        ),
    ]

    settings_panels = [
        MultiFieldPanel(
            [
                FieldRowPanel(
                    [
                        FieldPanel('form_golive_at'),
                        FieldPanel('form_expire_at'),
                    ],
                    classname='label-above',
                ),
            ],
            _('Form Scheduled Publishing'),
        ),
        FieldPanel('spam_protection')
    ]

    @property
    def form_live(self):
        """
        A boolean on whether or not the <form> element should be shown on the page.
        """
        return (self.form_golive_at is None or self.form_golive_at <= timezone.now()) and \
               (self.form_expire_at is None or self.form_expire_at >= timezone.now())

    def get_landing_page_template(self, request, *args, **kwargs):
        return self.landing_page_template

    def process_data(self, form, request):
        processed_data = {}
        # Handle file uploads
        for key, val in form.cleaned_data.items():

            if type(val) == InMemoryUploadedFile or type(val) == TemporaryUploadedFile:
                # Save the file and get its URL

                # Custom code to ensure that anonymous users get a session key.
                if not request.session.session_key:
                    request.session.create()

                directory = request.session.session_key
                storage = self.get_storage()
                Path(storage.path(directory)).mkdir(parents=True,
                                                    exist_ok=True)
                path = storage.get_available_name(
                    str(Path(directory) / val.name))
                with storage.open(path, 'wb+') as destination:
                    for chunk in val.chunks():
                        destination.write(chunk)

                processed_data[key] = "{0}{1}".format(cr_settings['PROTECTED_MEDIA_URL'], path)
            else:
                processed_data[key] = val

        return processed_data

    def get_storage(self):
        return FileSystemStorage(
            location=cr_settings['PROTECTED_MEDIA_ROOT'],
            base_url=cr_settings['PROTECTED_MEDIA_URL']
        )

    def process_form_submission(self, request, form, form_submission, processed_data):

        # Save to database
        if self.save_to_database:
            form_submission.save()

        # Send the mails
        if self.to_address:
            self.send_summary_mail(request, form, processed_data)

        if self.confirmation_emails:
            # Convert form data into a context.
            context = Context(self.data_to_dict(processed_data, request))
            # Render emails as if they are django templates.
            for email in self.confirmation_emails.all():

                # Build email message parameters.
                message_args = {}
                # From
                if email.from_address:
                    template_from_email = Template(email.from_address)
                    message_args['from_email'] = template_from_email.render(context)
                else:
                    genemail = GeneralSettings.for_request(request).from_email_address
                    if genemail:
                        message_args['from_email'] = genemail
                # Reply-to
                if email.reply_address:
                    template_reply_to = Template(email.reply_address)
                    message_args['reply_to'] = template_reply_to.render(context).split(',')
                # CC
                if email.cc_address:
                    template_cc = Template(email.cc_address)
                    message_args['cc'] = template_cc.render(context).split(',')
                # BCC
                if email.bcc_address:
                    template_bcc = Template(email.bcc_address)
                    message_args['bcc'] = template_bcc.render(context).split(',')
                # Subject
                if email.subject:
                    template_subject = Template(email.subject)
                    message_args['subject'] = template_subject.render(context)
                else:
                    message_args['subject'] = self.title
                # Body
                template_body = Template(email.body)
                message_args['body'] = template_body.render(context)
                # To
                template_to = Template(email.to_address)
                message_args['to'] = template_to.render(context).split(',')

                # Send email
                message = EmailMessage(**message_args)
                message.content_subtype = 'html'
                message.send()

        for fn in hooks.get_hooks('form_page_submit'):
            fn(instance=self, form_submission=form_submission)

    def send_summary_mail(self, request, form, processed_data):
        """
        Sends a form submission summary email.
        """
        addresses = [x.strip() for x in self.to_address.split(',')]
        content = []

        for key, value in self.data_to_dict(processed_data, request).items():
            content.append('{0}: {1}'.format(
                key.replace('_', ' ').title(),
                value
            ))

        content = '\n-------------------- \n'.join(content)

        # Build email message parameters
        message_args = {
            'body': content,
            'to': addresses,
        }
        if self.subject:
            message_args['subject'] = self.subject
        else:
            message_args['subject'] = self.title
        genemail = GeneralSettings.for_request(request).from_email_address
        if genemail:
            message_args['from_email'] = genemail
        if self.reply_address:
            # Render reply-to field using form submission as context.
            context = Context(self.data_to_dict(processed_data, request))
            template_reply_to = Template(self.reply_address)
            message_args['reply_to'] = template_reply_to.render(context).split(',')

        # Send email
        message = EmailMessage(**message_args)
        message.send()

    def render_landing_page(self, request, form_submission=None):

        """
        Renders the landing page.

        You can override this method to return a different HttpResponse as
        landing page. E.g. you could return a redirect to a separate page.
        """
        if self.thank_you_page:
            return redirect(self.thank_you_page.url)

        context = self.get_context(request)
        context['form_submission'] = form_submission
        response = render(
            request,
            self.get_landing_page_template(request),
            context
        )
        return response

    def data_to_dict(self, processed_data, request):
        """
        Converts processed form data into a dictionary suitable
        for rendering in a context.
        """
        dictionary = {}

        for key, value in processed_data.items():
            new_key = key.replace('-', '_')
            if isinstance(value, list):
                dictionary[new_key] = ', '.join(value)
            else:
                dictionary[new_key] = utils.attempt_protected_media_value_conversion(request, value)

        return dictionary

    preview_modes = [
        ('form', _('Form')),
        ('landing', _('Thank you page')),
    ]

    def serve_preview(self, request, mode):
        if mode == 'landing':
            request.is_preview = True
            return self.render_landing_page(request)

        return super().serve_preview(request, mode)

    def serve_submissions_list_view(self, request, *args, **kwargs):
        """
        Returns list submissions view for admin.

        `list_submissions_view_class` can be set to provide custom view class.
        Your class must be inherited from SubmissionsListView.
        """
        view = self.submissions_list_view_class.as_view()
        return view(request, form_page=self, *args, **kwargs)

    def get_form(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        form_params = self.get_form_parameters()
        form_params.update(kwargs)

        if request.method == 'POST':
            return form_class(request.POST, request.FILES, *args, **form_params)
        return form_class(*args, **form_params)

    def contains_spam(self, request):
        """
        Checks to see if the spam honeypot was filled out.
        """
        if request.POST.get("cr-decoy-comments", None):
            return True
        return False

    def process_spam_request(self, form, request):
        """
        Called when spam is found in the request.
        """
        messages.error(request, self.get_spam_message())
        logger.info("Detected spam submission on page: {0}\n{1}".format(self.title, vars(request)))

        return self.process_form_get(form, request)

    def get_spam_message(self):
        return _("There was an error while processing your submission.  Please try again.")

    def process_form_post(self, form, request):
        if form.is_valid():
            processed_data = self.process_data(form, request)
            form_submission = self.get_submission_class()(
                form_data=json.dumps(processed_data, cls=self.encoder),
                page=self,
            )
            self.process_form_submission(
                request=request,
                form=form,
                form_submission=form_submission,
                processed_data=processed_data)
            return self.render_landing_page(request, form_submission)
        return self.process_form_get(form, request)

    def process_form_get(self, form, request):
        context = self.get_context(request)
        context['form'] = form
        response = render(
            request,
            self.get_template(request),
            context
        )
        return response

    def serve(self, request, *args, **kwargs):
        form = self.get_form(request, page=self, user=request.user)
        if request.method == 'POST':
            if self.spam_protection and self.contains_spam(request):
                return self.process_spam_request(form, request)
            return self.process_form_post(form, request)
        return self.process_form_get(form, request)


class CoderedFormPage(CoderedFormMixin, CoderedWebPage):
    """
    This is basically a clone of wagtail.contrib.forms.models.AbstractForm
    with changes in functionality and extending CoderedWebPage vs wagtailcore.Page.
    """
    class Meta:
        verbose_name = _('CodeRed Form Page')
        abstract = True

    template = 'coderedcms/pages/form_page.html'
    landing_page_template = 'coderedcms/pages/form_page_landing.html'

    base_form_class = WagtailAdminFormPageForm

    form_builder = CoderedFormBuilder

    body_content_panels = [
        InlinePanel('form_fields', label="Form fields"),
    ] + \
        CoderedWebPage.body_content_panels + \
        CoderedFormMixin.body_content_panels + [
            FormSubmissionsPanel(),
            InlinePanel('confirmation_emails', label=_('Confirmation Emails'))
    ]

    settings_panels = CoderedPage.settings_panels + CoderedFormMixin.settings_panels

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not hasattr(self, 'landing_page_template'):
            name, ext = os.path.splitext(self.template)
            self.landing_page_template = name + '_landing' + ext

    def get_form_fields(self):
        """
        Form page expects `form_fields` to be declared.
        If you want to change backwards relation name,
        you need to override this method.
        """

        return self.form_fields.all()

    def get_data_fields(self):
        """
        Returns a list of tuples with (field_name, field_label).
        """

        data_fields = [
            ('submit_time', _('Submission date')),
        ]
        data_fields += [
            (field.clean_name, field.label)
            for field in self.get_form_fields()
        ]
        return data_fields

    def get_form_class(self):
        fb = self.form_builder(self.get_form_fields())
        return fb.get_form_class()

    def get_form_parameters(self):
        return {}

    def get_submission_class(self):
        """
        Returns submission class.

        You can override this method to provide custom submission class.
        Your class must be inherited from AbstractFormSubmission.
        """

        return FormSubmission


class CoderedSubmissionRevision(SubmissionRevision, models.Model):
    pass


class CoderedSessionFormSubmission(SessionFormSubmission):

    INCOMPLETE = 'incomplete'
    COMPLETE = 'complete'
    REVIEWED = 'reviewed'
    APPROVED = 'approved'
    REJECTED = 'rejected'
    STATUSES = (
        (INCOMPLETE, _('Not submitted')),
        (COMPLETE, _('Complete')),
        (REVIEWED, _('Under consideration')),
        (APPROVED, _('Approved')),
        (REJECTED, _('Rejected')),
    )
    status = models.CharField(max_length=10, choices=STATUSES, default=INCOMPLETE)

    def create_normal_submission(self, delete_self=True):
        submission_data = self.get_data()
        if 'user' in submission_data:
            submission_data['user'] = str(submission_data['user'])
        submission = FormSubmission.objects.create(
            form_data=json.dumps(submission_data, cls=StreamFormJSONEncoder),
            page=self.page
        )

        if delete_self:
            CoderedSubmissionRevision.objects.filter(submission_id=self.id).delete()
            self.delete()

        return submission

    def render_email(self, value):
        return value

    def render_link(self, value):
        return "{0}{1}".format(cr_settings['PROTECTED_MEDIA_URL'], value)

    def render_image(self, value):
        return "{0}{1}".format(cr_settings['PROTECTED_MEDIA_URL'], value)

    def render_file(self, value):
        return "{0}{1}".format(cr_settings['PROTECTED_MEDIA_URL'], value)


@receiver(post_save)
def create_submission_changed_revision(sender, **kwargs):
    if not issubclass(sender, SessionFormSubmission):
        return
    submission = kwargs['instance']
    created = kwargs['created']
    CoderedSubmissionRevision.create_from_submission(
        submission, (CoderedSubmissionRevision.CREATED if created else CoderedSubmissionRevision.CHANGED))  # noqa


@receiver(post_delete)
def create_submission_deleted_revision(sender, **kwargs):
    if not issubclass(sender, CoderedSessionFormSubmission):
        return
    submission = kwargs['instance']
    CoderedSubmissionRevision.create_from_submission(submission, SubmissionRevision.DELETED)  # noqa


class CoderedStep(Step):

    def get_markups_and_bound_fields(self, form):
        for struct_child in self.form_fields:
            block = struct_child.block
            if isinstance(block, FormFieldBlock):
                struct_value = struct_child.value
                field_name = block.get_slug(struct_value)
                yield form[field_name], 'field', struct_child
            else:
                yield mark_safe(struct_child), 'markup'


class CoderedSteps(Steps):

    def __init__(self, page, request=None):
        self.page = page
        # TODO: Make it possible to change the `form_fields` attribute.
        self.form_fields = page.form_fields
        self.request = request
        has_steps = any(isinstance(struct_child.block, FormStepBlock)
                        for struct_child in self.form_fields)
        if has_steps:
            steps = [CoderedStep(self, i, form_field)
                     for i, form_field in enumerate(self.form_fields)]
        else:
            steps = [CoderedStep(self, 0, self.form_fields)]
        super(Steps, self).__init__(steps)


class CoderedStreamFormMixin(StreamFormMixin):
    class Meta:
        abstract = True

    def get_steps(self, request=None):
        if not hasattr(self, 'steps'):
            steps = CoderedSteps(self, request=request)
            if request is None:
                return steps
            self.steps = steps
        return self.steps

    @staticmethod
    def get_submission_class():
        return FormSubmission

    @staticmethod
    def get_session_submission_class():
        return CoderedSessionFormSubmission

    def get_submission(self, request):
        Submission = self.get_session_submission_class()
        if request.user.is_authenticated:
            user_submission = Submission.objects.filter(
                user=request.user, page=self).order_by('-pk').first()
            if user_submission is None:
                return Submission(user=request.user, page=self, form_data='[]')
            return user_submission

        # Custom code to ensure that anonymous users get a session key.
        if not request.session.session_key:
            request.session.create()

        user_submission = Submission.objects.filter(
            session_key=request.session.session_key, page=self
        ).order_by('-pk').first()
        if user_submission is None:
            return Submission(session_key=request.session.session_key,
                              page=self, form_data='[]')
        return user_submission


class CoderedStreamFormPage(CoderedFormMixin, CoderedStreamFormMixin, CoderedWebPage):
    class Meta:
        verbose_name = _('CodeRed Advanced Form Page')
        abstract = True

    template = 'coderedcms/pages/stream_form_page.html'
    landing_page_template = 'coderedcms/pages/form_page_landing.html'

    form_fields = StreamField(STREAMFORM_BLOCKS)
    encoder = StreamFormJSONEncoder

    body_content_panels = [
        StreamFieldPanel('form_fields')
    ] + \
        CoderedFormMixin.body_content_panels + [
            InlinePanel('confirmation_emails', label=_('Confirmation Emails'))
    ]

    def process_form_post(self, form, request):
        if form.is_valid():
            is_complete = self.steps.update_data()
            if is_complete:
                submission = self.get_submission(request)
                self.process_form_submission(
                    request=request,
                    form=form,
                    form_submission=submission,
                    processed_data=submission.get_data()
                )
                normal_submission = submission.create_normal_submission()
                return self.render_landing_page(request, normal_submission)
            return HttpResponseRedirect(self.url)
        return self.process_form_get(form, request)

    def process_form_get(self, form, request):
        return CoderedWebPage.serve(self, request)

    def get_form(self, request, *args, **kwargs):
        return self.get_context(request)['form']

    def get_storage(self):
        return FileSystemStorage(
            location=cr_settings['PROTECTED_MEDIA_ROOT'],
            base_url=cr_settings['PROTECTED_MEDIA_URL']
        )


class CoderedLocationPage(CoderedWebPage):
    """
    Location, suitable for store locations or help centers.
    """
    class Meta:
        verbose_name = _('CodeRed Location')
        abstract = True

    template = 'coderedcms/pages/location_page.html'

    # Override body to provide simpler content
    body = StreamField(CONTENT_STREAMBLOCKS, null=True, blank=True)

    address = models.TextField(
        blank=True,
        verbose_name=_("Address")
    )
    latitude = models.FloatField(
        blank=True,
        null=True,
        verbose_name=_("Latitude")
    )
    longitude = models.FloatField(
        blank=True,
        null=True,
        verbose_name=_("Longitude")
    )
    auto_update_latlng = models.BooleanField(
        default=True,
        verbose_name=_("Auto Update Latitude and Longitude"),
        help_text=_("If checked, automatically update the latitude and longitude when the address is updated.")  # noqa
    )
    map_title = models.CharField(
        blank=True,
        max_length=255,
        verbose_name=_("Map Title"),
        help_text=_("If this is filled out, this is the title that will be used on the map.")
    )
    map_description = models.CharField(
        blank=True,
        max_length=255,
        verbose_name=_("Map Description"),
        help_text=_("If this is filled out, this is the description that will be used on the map.")
    )
    website = models.TextField(
        blank=True,
        verbose_name=_("Website")
    )
    phone_number = models.CharField(
        blank=True,
        max_length=255,
        verbose_name=_("Phone Number")
    )

    content_panels = CoderedWebPage.content_panels + [
        FieldPanel('address'),
        FieldPanel('website'),
        FieldPanel('phone_number'),
    ]

    layout_panels = CoderedWebPage.layout_panels + [
        MultiFieldPanel(
            [
                FieldPanel('map_title'),
                FieldPanel('map_description'),
            ],
            heading=_('Map Layout')
        ),
    ]

    settings_panels = CoderedWebPage.settings_panels + [
        MultiFieldPanel(
            [
                FieldPanel('auto_update_latlng'),
                FieldPanel('latitude'),
                FieldPanel('longitude'),
            ],
            heading=_("Location Settings")
        ),
    ]

    @property
    def geojson_name(self):
        return self.map_title or self.title

    @property
    def geojson_description(self):
        return self.map_description

    @property
    def render_pin_description(self):
        return render_to_string(
            'coderedcms/includes/map_pin_description.html',
            {
                'page': self
            }
        )

    @property
    def render_list_description(self):
        return render_to_string(
            'coderedcms/includes/map_list_description.html',
            {
                'page': self
            }
        )

    def to_geojson(self):
        return {
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": [self.longitude, self.latitude]
            },
            "properties": {
                "list_description": self.render_list_description,
                "pin_description": self.render_pin_description
            }
        }

    def save(self, *args, **kwargs):
        if self.auto_update_latlng and GoogleApiSettings.for_site(
            Site.objects.get(is_default_site=True)
        ).google_maps_api_key:
            try:
                g = geocoder.google(self.address, key=GoogleApiSettings.for_site(
                    Site.objects.get(is_default_site=True)
                ).google_maps_api_key)
                self.latitude = g.latlng[0]
                self.longitude = g.latlng[1]
            except TypeError:
                # Raised if google denied the request
                pass

        return super(CoderedLocationPage, self).save(*args, **kwargs)

    def get_context(self, request, *args, **kwargs):
        context = super().get_context(request)
        context['google_api_key'] = GoogleApiSettings.for_site(
            Site.objects.get(is_default_site=True)
        ).google_maps_api_key
        return context


class CoderedLocationIndexPage(CoderedWebPage):
    """
    Shows a map view of the children CoderedLocationPage.
    """
    class Meta:
        verbose_name = _('CodeRed Location Index Page')
        abstract = True

    template = 'coderedcms/pages/location_index_page.html'

    index_show_subpages_default = True

    center_latitude = models.FloatField(
        null=True,
        blank=True,
        help_text=_('The default latitude you want the map set to.'),
        default=0
    )
    center_longitude = models.FloatField(
        null=True,
        blank=True,
        help_text=_('The default longitude you want the map set to.'),
        default=0
    )
    zoom = models.IntegerField(
        default=8,
        validators=[
            MaxValueValidator(20),
            MinValueValidator(1),
        ],
        help_text=_("Requires API key to use zoom. 1: World, 5: Landmass/continent, 10: City, 15: Streets, 20: Buildings")  # noqa
    )

    layout_panels = CoderedWebPage.layout_panels + [
        MultiFieldPanel(
            [
                FieldPanel('center_latitude'),
                FieldPanel('center_longitude'),
                FieldPanel('zoom'),
            ],
            heading=_('Map Display')
        ),
    ]

    def geojson_data(self, viewport=None):
        """
        function that will return all locations under this index as geoJSON compliant data.
        It is filtered by a latitude/longitude viewport if given.

        viewport is a string in the format of :
        'southwest.latitude,southwest.longitude|northeast.latitude,northeast.longitude'

        An example viewport that covers Cleveland, OH would look like this:
        '41.354912150983964,-81.95331736661791|41.663427748126935,-81.45206614591478'
        """
        qs = self.get_index_children().live()

        if viewport:
            southwest, northeast = viewport.split('|')
            southwest = [float(x) for x in southwest.split(',')]
            northeast = [float(x) for x in northeast.split(',')]

            qs = qs.filter(
                latitude__gte=southwest[0],
                latitude__lte=northeast[0],
                longitude__gte=southwest[1],
                longitude__lte=northeast[1]
            )

        return {
            "type": "FeatureCollection",
            "features": [
                location.to_geojson() for location in qs
            ]
        }

    def serve(self, request, *args, **kwargs):
        data_format = request.GET.get('data-format', None)
        if data_format == 'geojson':
            return self.serve_geojson(request, *args, **kwargs)
        return super().serve(request, *args, **kwargs)

    def serve_geojson(self, request, *args, **kwargs):
        viewport = request.GET.get('viewport', None)
        return JsonResponse(self.geojson_data(viewport=viewport))

    def get_context(self, request, *args, **kwargs):
        context = super().get_context(request)
        context['google_api_key'] = GoogleApiSettings.for_site(
            Site.objects.get(is_default_site=True)
        ).google_maps_api_key
        return context



# Generated by Django 3.1.13 on 2021-07-30 20:37

import coderedcms.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('coderedcms', '0020_auto_20210527_1030'),
    ]

    operations = [
        migrations.AddField(
            model_name='analyticssettings',
            name='body_scripts',
            field=coderedcms.fields.MonospaceField(blank=True, help_text='Add tracking scripts toward closing <body> tag.', null=True, verbose_name='<body> tracking scripts'),
        ),
        migrations.AddField(
            model_name='analyticssettings',
            name='head_scripts',
            field=coderedcms.fields.MonospaceField(blank=True, help_text='Add tracking scripts between the <head> tags.', null=True, verbose_name='<head> tracking scripts'),
        ),
    ]



from collections import OrderedDict
from importlib import import_module
from itertools import zip_longest
import json
import os
from pathlib import Path

from PIL import Image
import datetime
from django.apps import apps
from django.conf import settings
from django.contrib import messages
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.files.storage import default_storage
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import (
    CharField, TextField, DateTimeField, Model, ForeignKey, PROTECT, CASCADE,
    QuerySet,
)
from django.db.models.fields.files import FieldFile
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.forms import Form, ImageField, FileField, URLField, EmailField
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.utils.safestring import SafeData, mark_safe
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from wagtail.core.models import Page
from wagtail.contrib.forms.models import (
    AbstractForm, AbstractEmailForm, AbstractFormSubmission)

from .blocks import FormStepBlock, FormFieldBlock


class Step:
    def __init__(self, steps, index, struct_child):
        self.steps = steps
        self.index = index
        block = getattr(struct_child, 'block', None)
        if block is None:
            struct_child = []
        if isinstance(block, FormStepBlock):
            self.name = struct_child.value['name']
            self.form_fields = struct_child.value['form_fields']
        else:
            self.name = ''
            self.form_fields = struct_child

    @property
    def index1(self):
        return self.index + 1

    @property
    def url(self):
        return '%s?step=%s' % (self.steps.page.url, self.index1)

    def get_form_fields(self):
        form_fields = OrderedDict()
        field_blocks = self.form_fields
        for struct_child in field_blocks:
            block = struct_child.block
            if isinstance(block, FormFieldBlock):
                struct_value = struct_child.value
                field_name = block.get_slug(struct_value)
                form_fields[field_name] = block.get_field(struct_value)
        return form_fields

    def get_form_class(self):
        return type('WagtailForm', self.steps.page.get_form_class_bases(),
                    self.get_form_fields())

    def get_markups_and_bound_fields(self, form):
        for struct_child in self.form_fields:
            block = struct_child.block
            if isinstance(block, FormFieldBlock):
                struct_value = struct_child.value
                field_name = block.get_slug(struct_value)
                yield form[field_name], 'field'
            else:
                yield mark_safe(struct_child), 'markup'

    def __str__(self):
        if self.name:
            return self.name
        return _('Step %s') % self.index1

    @property
    def badge(self):
        return (mark_safe('<span class="badge">%s/%s</span>')
                % (self.index1, len(self.steps)))

    def __html__(self):
        return '%s %s' % (self, self.badge)

    @property
    def is_active(self):
        return self.index == self.steps.current_index

    @property
    def is_last(self):
        return self.index1 == len(self.steps)

    @property
    def has_prev(self):
        return self.index > 0

    @property
    def has_next(self):
        return self.index1 < len(self.steps)

    @property
    def prev(self):
        if self.has_prev:
            return self.steps[self.index-1]

    @property
    def next(self):
        if self.has_next:
            return self.steps[self.index+1]

    def get_existing_data(self, raw=False):
        data = self.steps.get_existing_data()[self.index]
        fields = self.get_form_fields()
        if not raw:
            class FakeField:
                storage = self.steps.get_storage()

            for field_name, value in data.items():
                if field_name in fields and isinstance(fields[field_name],
                                                       FileField):
                    data[field_name] = FieldFile(None, FakeField, value)
        return data

    @property
    def is_available(self):
        return self.prev is None or self.prev.get_existing_data(raw=True)


class StreamFormJSONEncoder(DjangoJSONEncoder):
    def default(self, o):
        try:
            from phonenumber_field.phonenumber import PhoneNumber
        except ImportError:
            pass
        else:
            if isinstance(o, PhoneNumber):
                return str(o)

        return super().default(o)


class Steps(list):
    def __init__(self, page, request=None):
        self.page = page
        # TODO: Make it possible to change the `form_fields` attribute.
        self.form_fields = page.form_fields
        self.request = request
        has_steps = any(isinstance(struct_child.block, FormStepBlock)
                        for struct_child in self.form_fields)
        if has_steps:
            steps = [Step(self, i, form_field)
                     for i, form_field in enumerate(self.form_fields)]
        else:
            steps = [Step(self, 0, self.form_fields)]
        super().__init__(steps)

    def clamp_index(self, index: int):
        if index < 0:
            index = 0
        if index >= len(self):
            index = len(self) - 1
        while not self[index].is_available:
            index -= 1
        return index

    @property
    def current_index(self):
        return self.request.session.get(self.page.current_step_session_key, 0)

    @property
    def current(self):
        return self[self.current_index]

    @current.setter
    def current(self, new_index: int):
        if not isinstance(new_index, int):
            raise TypeError('Use an integer to set the new current step.')
        self.request.session[self.page.current_step_session_key] = \
            self.clamp_index(new_index)

    def forward(self, increment: int = 1):
        self.current = self.current_index + increment

    def backward(self, increment: int = 1):
        self.current = self.current_index - increment

    def get_submission(self):
        return self.page.get_submission(self.request)

    def get_existing_data(self):
        submission = self.get_submission()
        data = [] if submission is None else json.loads(submission.form_data)
        length_difference = len(self) - len(data)
        if length_difference > 0:
            data.extend([{}] * length_difference)
        return data

    def get_current_form(self):
        request = self.request
        if request.method == 'POST':
            step_value = request.POST.get('step', 'next')
            if step_value == 'prev':
                self.backward()
            else:
                return self.current.get_form_class()(
                    request.POST, request.FILES,
                    initial=self.current.get_existing_data())
        return self.current.get_form_class()(
            initial=self.current.get_existing_data())

    def get_storage(self):
        return self.page.get_storage()

    def save_files(self, form):
        submission = self.get_submission()
        for name, field in form.fields.items():
            if isinstance(field, FileField):
                file = form.cleaned_data[name]
                if file == form.initial.get(name, ''):  # Nothing submitted.
                    form.cleaned_data[name] = file.name
                    continue
                if submission is not None:
                    submission.delete_file(name)
                if not file:  # 'Clear' was checked.
                    form.cleaned_data[name] = ''
                    continue
                directory = self.request.session.session_key
                storage = self.get_storage()
                Path(storage.path(directory)).mkdir(parents=True,
                                                    exist_ok=True)
                path = storage.get_available_name(
                    str(Path(directory) / file.name))
                with storage.open(path, 'wb+') as destination:
                    for chunk in file.chunks():
                        destination.write(chunk)
                form.cleaned_data[name] = path

    def update_data(self):
        form = self.get_current_form()
        if form.is_valid():
            form_data = self.get_existing_data()
            self.save_files(form)
            form_data[self.current_index] = form.cleaned_data
            form_data = json.dumps(form_data, cls=StreamFormJSONEncoder)
            is_complete = self.current.is_last
            submission = self.get_submission()
            submission.form_data = form_data
            if not submission.is_complete and is_complete:
                submission.status = submission.COMPLETE
            submission.save()
            if is_complete:
                self.current = 0
            else:
                self.forward()
            return is_complete
        return False


class SessionFormSubmission(AbstractFormSubmission):

    session_key = CharField(max_length=40, null=True, default=None)
    user = ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True,
                      related_name='+', on_delete=PROTECT)
    thumbnails_by_path = TextField(default=json.dumps({}))
    last_modification = DateTimeField(_('last modification'), auto_now=True)
    INCOMPLETE = 'incomplete'
    COMPLETE = 'complete'
    REVIEWED = 'reviewed'
    APPROVED = 'approved'
    REJECTED = 'rejected'
    STATUSES = (
        (INCOMPLETE, _('Not submitted')),
        (COMPLETE, _('In progress')),
        (REVIEWED, _('Under consideration')),
        (APPROVED, _('Approved')),
        (REJECTED, _('Rejected')),
    )
    status = CharField(max_length=10, choices=STATUSES, default=INCOMPLETE)

    class Meta:
        verbose_name = _('form submission')
        verbose_name_plural = _('form submissions')
        unique_together = (('page', 'session_key'),
                           ('page', 'user'))
        abstract = True

    @property
    def is_complete(self):
        return self.status != self.INCOMPLETE

    @property
    def form_page(self):
        return self.page.specific

    def get_session(self):
        return import_module(settings.SESSION_ENGINE).SessionStore(
            session_key=self.session_key)

    def reset_step(self):
        session = self.get_session()
        try:
            del session[self.form_page.current_step_session_key]
        except KeyError:
            pass
        else:
            session.save()

    def get_storage(self):
        return self.form_page.get_storage()

    def get_thumbnail_path(self, path, width=64, height=64):
        if not path:
            return ''
        variant = '%s×%s' % (width, height)
        thumbnails_by_path = json.loads(self.thumbnails_by_path)
        thumbnails_paths = thumbnails_by_path.get(path)
        if thumbnails_paths is None:
            thumbnails_by_path[path] = {}
        else:
            thumbnail_path = thumbnails_paths.get(variant)
            if thumbnail_path is not None:
                return thumbnail_path

        path = Path(path)
        thumbnail_path = str(path.with_suffix('.%s%s'
                                              % (variant, path.suffix)))
        storage = self.get_storage()
        thumbnail_path = storage.get_available_name(thumbnail_path)

        thumbnail = Image.open(storage.path(path))
        thumbnail.thumbnail((width, height))
        thumbnail.save(storage.path(thumbnail_path))

        thumbnails_by_path[str(path)][variant] = thumbnail_path
        self.thumbnails_by_path = json.dumps(thumbnails_by_path,
                                             cls=StreamFormJSONEncoder)
        self.save()
        return thumbnail_path

    def get_fields(self, by_step=False):
        return self.form_page.get_form_fields(by_step=by_step)

    def get_existing_thumbnails(self, path):
        thumbnails_paths = json.loads(self.thumbnails_by_path).get(path, {})
        for thumbnail_path in thumbnails_paths.values():
            yield thumbnail_path

    def get_files_by_field(self):
        data = self.get_data(raw=True)
        files = {}
        for name, field in self.get_fields().items():
            if isinstance(field, FileField):
                path = data.get(name)
                if path:
                    files[name] = [path] + list(
                        self.get_existing_thumbnails(path))
        return files

    def get_all_files(self):
        for paths in self.get_files_by_field().values():
            for path in paths:
                yield path

    def delete_file(self, field_name):
        thumbnails_by_path = json.loads(self.thumbnails_by_path)
        for path in self.get_files_by_field().get(field_name, ()):
            self.get_storage().delete(path)
            if path in thumbnails_by_path:
                del thumbnails_by_path[path]
        self.thumbnails_by_path = json.dumps(thumbnails_by_path,
                                             cls=StreamFormJSONEncoder)
        self.save()

    def render_email(self, value):
        return (mark_safe('<a href="mailto:%s" target="_blank">%s</a>')
                % (value, value))

    def render_link(self, value):
        return (mark_safe('<a href="%s" target="_blank">%s</a>')
                % (value, value))

    def render_image(self, value):
        storage = self.get_storage()
        return (mark_safe('<a href="%s" target="_blank"><img src="%s" /></a>')
                % (storage.url(value),
                   storage.url(self.get_thumbnail_path(value))))

    def render_file(self, value):
        return mark_safe('<a href="%s" target="_blank">%s</a>') % (
            self.get_storage().url(value),
            Path(value).name
        )

    def format_value(self, field, value):
        if value is None or value == '':
            return '-'
        new_value = self.form_page.format_value(field, value)
        if new_value != value:
            return new_value
        if value is True:
            return 'Yes'
        if value is False:
            return 'No'
        if isinstance(value, (list, tuple)):
            return ', '.join([self.format_value(field, item)
                              for item in value])
        if isinstance(value, datetime.date):
            return value
        if isinstance(field, EmailField):
            return self.render_email(value)
        if isinstance(field, URLField):
            return self.render_link(value)
        if isinstance(field, ImageField):
            return self.render_image(value)
        if isinstance(field, FileField):
            return self.render_file(value)
        if isinstance(value, SafeData) or hasattr(value, '__html__'):
            return value
        return str(value)

    def format_db_field(self, field_name, raw=False):
        method = getattr(self, 'get_%s_display' % field_name, None)
        if method is not None:
            return method()
        value = getattr(self, field_name)
        if raw:
            return value
        return self.format_value(self._meta.get_field(field_name).formfield(),
                                 value)

    def get_steps_data(self, raw=False):
        steps_data = json.loads(self.form_data)
        if raw:
            return steps_data
        fields_and_data_iterator = zip_longest(self.get_fields(by_step=True),
                                               steps_data, fillvalue={})
        return [
            OrderedDict([(name, self.format_value(field, step_data.get(name)))
                         for name, field in step_fields.items()])
            for step_fields, step_data in fields_and_data_iterator]

    def get_extra_data(self, raw=False):
        return self.form_page.get_extra_data(self, raw=raw)

    def get_data(self, raw=False, add_metadata=True):
        steps_data = self.get_steps_data(raw=raw)
        form_data = {}
        form_data.update(self.get_extra_data(raw=raw))
        for step_data in steps_data:
            form_data.update(step_data)
        if add_metadata:
            form_data.update(
                status=self.format_db_field('status', raw=raw),
                user=self.format_db_field('user', raw=raw),
                submit_time=self.format_db_field('submit_time', raw=raw),
                last_modification=self.format_db_field('last_modification',
                                                       raw=raw),
            )
        return form_data

    def steps_with_data_iterator(self, raw=False):
        for step, step_data_fields, step_data in zip(
                self.form_page.get_steps(),
                self.form_page.get_data_fields(by_step=True),
                self.get_steps_data(raw=raw)):
            yield step, [(field_name, field_label, step_data[field_name])
                         for field_name, field_label in step_data_fields]


@receiver(post_delete, sender=SessionFormSubmission)
def delete_files(sender, **kwargs):
    instance = kwargs['instance']
    instance.reset_step()
    storage = instance.get_storage()
    for path in instance.get_all_files():
        storage.delete(path)

        # Automatically deletes ancestor folders if empty.
        directory = Path(path)
        while directory.parent != Path(directory.root):
            directory = directory.parent
            try:
                subdirectories, files = storage.listdir(directory)
            except FileNotFoundError:
                continue
            if not subdirectories and not files:
                Path(storage.path(directory)).rmdir()


class SubmissionRevisionQuerySet(QuerySet):
    def for_submission(self, submission):
        return self.filter(**self.model.get_filters_for(submission))

    def created(self):
        return self.filter(type=self.model.CREATED)

    def changed(self):
        return self.filter(type=self.model.CHANGED)

    def deleted(self):
        return self.filter(type=self.model.DELETED)


class SubmissionRevision(Model):
    CREATED = 'created'
    CHANGED = 'changed'
    DELETED = 'deleted'
    TYPES = (
        (CREATED, _('Created')),
        (CHANGED, _('Changed')),
        (DELETED, _('Deleted')),
    )
    type = CharField(max_length=7, choices=TYPES)
    created_at = DateTimeField(auto_now_add=True)
    submission_ct = ForeignKey('contenttypes.ContentType', on_delete=CASCADE)
    submission_id = TextField()
    submission = GenericForeignKey('submission_ct', 'submission_id')
    data = TextField()
    summary = TextField()

    objects = SubmissionRevisionQuerySet.as_manager()

    class Meta:
        ordering = ('-created_at',)
        abstract = True

    @staticmethod
    def get_filters_for(submission):
        return {
            'submission_ct':
                ContentType.objects.get_for_model(submission._meta.model),
            'submission_id': str(submission.pk),
        }

    @classmethod
    def diff_summary(cls, page, data1, data2):
        diff = []
        data_fields = page.get_data_fields()
        hidden_types = (tuple, list, dict)
        for k, label in data_fields:
            value1 = data1.get(k)
            value2 = data2.get(k)
            if value2 == value1 or not value1 and not value2:
                continue
            is_hidden = (isinstance(value1, hidden_types)
                         or isinstance(value2, hidden_types))

            # Escapes newlines as they are used as separator inside summaries.
            if isinstance(value1, str):
                value1 = value1.replace('\n', r'\n')
            if isinstance(value2, str):
                value2 = value2.replace('\n', r'\n')

            if value2 and not value1:
                diff.append(
                    ((_('“%s” set.') % label) if is_hidden
                     else (_('“%s” set to “%s”.')) % (label, value2)))
            elif value1 and not value2:
                diff.append(_('“%s” unset.') % label)
            else:
                diff.append(((_('“%s” changed.') % label) if is_hidden
                             else (_('“%s” changed from “%s” to “%s”.')
                                   % (label, value1, value2))))
        return '\n'.join(diff)

    @classmethod
    def create_from_submission(cls, submission, revision_type):
        page = submission.form_page
        try:
            previous = cls.objects.for_submission(
                submission).latest('created_at')
        except cls.DoesNotExist:
            previous_data = {}
        else:
            previous_data = previous.get_data()
        filters = cls.get_filters_for(submission)
        data = submission.get_data(raw=True, add_metadata=False)
        data['status'] = submission.status
        if revision_type == cls.CREATED:
            summary = _('Submission created.')
        elif revision_type == cls.DELETED:
            summary = _('Submission deleted.')
        else:
            summary = cls.diff_summary(page, previous_data, data)
        if not summary:  # Nothing changed.
            return
        filters.update(
            type=revision_type, data=json.dumps(data, cls=StreamFormJSONEncoder), summary=summary
        )
        return cls.objects.create(**filters)

    def get_data(self):
        return json.loads(self.data)

# ORIGINAL NORIPYT CODE.
# We don't want these receivers triggering.

# @receiver(post_save)
# def create_submission_changed_revision(sender, **kwargs):
#     if not issubclass(sender, SessionFormSubmission):
#         return
#     submission = kwargs['instance']
#     created = kwargs['created']
#     SubmissionRevision.create_from_submission(
#         submission, (SubmissionRevision.CREATED if created
#                      else SubmissionRevision.CHANGED))


# @receiver(post_delete)
# def create_submission_deleted_revision(sender, **kwargs):
#     if not issubclass(sender, SessionFormSubmission):
#         return
#     submission = kwargs['instance']
#     SubmissionRevision.create_from_submission(submission,
#                                               SubmissionRevision.DELETED)


class StreamFormMixin:
    preview_modes = Page.DEFAULT_PREVIEW_MODES

    @property
    def current_step_session_key(self):
        return '%s:step' % self.pk

    def get_steps(self, request=None):
        if not hasattr(self, 'steps'):
            steps = Steps(self, request=request)
            if request is None:
                return steps
            self.steps = steps
        return self.steps

    def get_form_fields(self, by_step=False):
        if by_step:
            return [step.get_form_fields() for step in self.get_steps()]
        form_fields = OrderedDict()
        for step_fields in self.get_form_fields(by_step=True):
            form_fields.update(step_fields)
        return form_fields

    def get_context(self, request, *args, **kwargs):
        context = super().get_context(request, *args, **kwargs)
        self.steps = self.get_steps(request)
        step_value = request.GET.get('step')
        if step_value is not None and step_value.isdigit():
            self.steps.current = int(step_value) - 1
        form = self.steps.get_current_form()
        context.update(
            steps=self.steps,
            step=self.steps.current,
            form=form,
            markups_and_bound_fields=list(
                self.steps.current.get_markups_and_bound_fields(form)),
        )
        return context

    def get_storage(self):
        return default_storage

    @staticmethod
    def get_form_class_bases():
        return Form,

    @staticmethod
    def get_submission_class():
        return SessionFormSubmission

    def get_submission(self, request):
        Submission = self.get_submission_class()
        if request.user.is_authenticated:
            user_submission = Submission.objects.filter(
                user=request.user, page=self).order_by('-pk').first()
            if user_submission is None:
                return Submission(user=request.user, page=self, form_data='[]')
            return user_submission

        user_submission = Submission.objects.filter(
            session_key=request.session.session_key, page=self
        ).order_by('-pk').first()
        if user_submission is None:
            return Submission(session_key=request.session.session_key,
                              page=self, form_data='[]')
        return user_submission

    def get_success_url(self):
        form_complete_models = [model for model in apps.get_models()
                                if issubclass(model, FormCompleteMixin)]
        cts = (ContentType.objects
               .get_for_models(*form_complete_models).values())
        first_child = self.get_children().filter(content_type__in=cts).first()
        if first_child is None:
            return self.url
        return first_child.url

    def serve_success(self, request, *args, **kwargs):
        url = self.get_success_url()
        if url == self.url:
            messages.success(request,
                             _('Successfully submitted the form.'))
        return HttpResponseRedirect(url)

    def serve(self, request, *args, **kwargs):
        context = self.get_context(request)
        form = context['form']
        if request.method == 'POST' and form.is_valid():
            is_complete = self.steps.update_data()
            if is_complete:
                return self.serve_success(request, *args, **kwargs)
            return HttpResponseRedirect(self.url)
        return Page.serve(self, request, *args, **kwargs)

    def get_data_fields(self, by_step=False, add_metadata=True):
        if by_step:
            return [[(field_name, field.label)
                     for field_name, field in step_fields.items()]
                    for step_fields in self.get_form_fields(by_step=True)]

        data_fields = []
        data_fields.extend(self.get_extra_data_fields())
        if add_metadata:
            data_fields.extend((
                ('status', _('Status')),
                ('user', _('User')),
                ('submit_time', _('First modification')),
                ('last_modification', _('Last modification'))))
        data_fields.extend([
            (field_name, field_label)
            for step_data_fields in self.get_data_fields(by_step=True)
            for field_name, field_label in step_data_fields])
        return data_fields

    def get_extra_data_fields(self):
        return ()

    def get_extra_data(self, submission, raw=False):
        return {}

    def format_value(self, field, value):
        return value


class ClosingFormMixin(Model):
    closing_at = DateTimeField()

    closed_template = None

    class Meta:
        abstract = True

    @property
    def is_closed(self):
        return now() > self.closing_at

    def get_closed_template(self, request, *args, **kwargs):
        if self.closed_template is None:
            template = self.get_template(request, *args, **kwargs)
            base, ext = os.path.splitext(template)
            return '%s_closed%s' % (base, ext)
        return self.closed_template

    def serve_closed(self, request, *args, **kwargs):
        return TemplateResponse(
            request,
            self.get_closed_template(request, *args, **kwargs),
            self.get_context(request, *args, **kwargs),
        )

    def serve(self, request, *args, **kwargs):
        if self.is_closed:
            return self.serve_closed(request, *args, **kwargs)
        return super().serve(request, *args, **kwargs)


class FormCompleteMixin:
    def get_form_page(self):
        return self.get_parent().specific

    def serve(self, request, *args, **kwargs):
        form_page = self.get_form_page()
        if isinstance(form_page, LoginRequiredMixin) \
                and not request.user.is_authenticated():
            return HttpResponseRedirect(form_page.url)
        self.submission = form_page.get_submission(request)
        if self.submission is not None and self.submission.is_complete \
                or getattr(request, 'is_preview', False):
            return super().serve(request, *args, **kwargs)
        return HttpResponseRedirect(form_page.url)

    def get_context(self, *args, **kwargs):
        context = super().get_context(*args, **kwargs)
        if hasattr(self, 'submission'):
            context['submission'] = self.submission
        return context


class LoginRequiredMixin:
    login_required_template = None

    def get_login_required_template(self, request, *args, **kwargs):
        if self.login_required_template is None:
            template = self.get_template(request, *args, **kwargs)
            base, ext = os.path.splitext(template)
            return '%s_login_required%s' % (base, ext)
        return self.login_required_template

    def serve_login_required(self, request, *args, **kwargs):
        return TemplateResponse(
            request,
            self.get_login_required_template(request, *args, **kwargs),
            self.get_context(request, *args, **kwargs),
        )

    def serve(self, request, *args, **kwargs):
        if not request.user.is_authenticated():
            return self.serve_login_required(request, *args, **kwargs)
        return super().serve(request, *args, **kwargs)


class AbstractStreamForm(StreamFormMixin, AbstractForm):
    class Meta:
        abstract = True


class AbstractEmailStreamForm(StreamFormMixin, AbstractEmailForm):
    class Meta:
        abstract = True




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django import VERSION as DJANGO_VERSION
from django.db import migrations

from wagtail.core.models import Locale


def initial_data(apps, schema_editor):
    ContentType = apps.get_model('contenttypes.ContentType')
    Page = apps.get_model('wagtailcore.Page')
    Site = apps.get_model('wagtailcore.Site')
    WebPage = apps.get_model('testapp.WebPage')

    # Create page content type
    webpage_content_type, created = ContentType.objects.get_or_create(
        model='webpage',
        app_label='testapp',
    )

    # Delete the default home page generated by wagtail,
    # and replace it with a more useful page type.
    curr_homepage = Page.objects.filter(slug='home').delete()

    homepage = WebPage.objects.create(
        title = "Home",
        slug='home',
        custom_template='coderedcms/pages/home_page.html',
        content_type=webpage_content_type,
        path='00010001',
        depth=2,
        numchild=0,
        url_path='/home/',
        locale_id=Locale.get_default().id,
    )

    # Create a new default site
    Site.objects.create(
        hostname='',
        site_name='Test Site',
        root_page_id=homepage.id,
        is_default_site=True
    )


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0001_initial'),
        ('wagtailcore', '0002_initial_data'),
    ]

    operations = [
        migrations.RunPython(initial_data),
    ]



# Generated by Django 2.0.9 on 2018-11-19 20:07

import coderedcms.blocks.base_blocks
from django.db import migrations, models
import django.db.models.deletion
import wagtail.contrib.table_block.blocks
import wagtail.core.blocks
import wagtail.core.fields
import wagtail.documents.blocks
import wagtail.images.blocks
import wagtail.snippets.blocks


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0040_page_draft_title'),
        ('coderedcms', '0003_auto_20180912_1632'),
    ]

    operations = [
        migrations.CreateModel(
            name='GoogleApiSettings',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('google_maps_api_key', models.CharField(blank=True, help_text='The API Key used for Google Maps.', max_length=255, verbose_name='Google Maps API Key')),
                ('site', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.Site')),
            ],
            options={
                'verbose_name': 'Google API Settings',
            },
        ),
    ]



"""
Maintains version of coderedcms.
4th element is for pre-releases. Leave blank for stable releases.

    X.Y.ZaN   # Alpha release
    X.Y.ZbN   # Beta release
    X.Y.ZrcN  # Release Candidate
    X.Y.Z     # Final release

5th element is for dev/post releases. Leave blank for stable releases.

    X.Y.Z.devN   # Development release
    X.Y.Z.postN  # Post release (e.g. docs changes but no actual code changes)

See: https://www.python.org/dev/peps/pep-0440/
"""
release = ["0", "21", "1", "", ""]


def _get_version() -> str:
    v = "{0}.{1}.{2}".format(release[0], release[1], release[2])
    if len(release) >= 4 and release[3]:
        v = "{0}{1}".format(v, release[3])
    if len(release) >= 5 and release[4]:
        v = "{0}.{1}".format(v, release[4])
    return v


__version__ = _get_version()
__shortversion__ = "{0}.{1}".format(release[0], release[1])



import mimetypes
import os
from itertools import chain
from datetime import datetime
from django.http import Http404, HttpResponse, HttpResponsePermanentRedirect, JsonResponse
from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.core.paginator import Paginator, InvalidPage, EmptyPage, PageNotAnInteger
from django.shortcuts import redirect, render
from django.utils.translation import ungettext, gettext_lazy as _
from icalendar import Calendar
from wagtail.admin import messages
from wagtail.search.backends import db, get_search_backend
from wagtail.search.models import Query
from coderedcms import utils
from coderedcms.forms import SearchForm
from coderedcms.models import (
    CoderedPage,
    CoderedEventPage,
    get_page_models,
    GeneralSettings,
    LayoutSettings
)
from coderedcms.importexport import convert_csv_to_json, import_pages, ImportPagesFromCSVFileForm
from coderedcms.settings import cr_settings


def search(request):
    """
    Searches pages across the entire site.
    """
    search_form = SearchForm(request.GET)
    pagetypes = []
    results = None
    results_paginated = None

    if search_form.is_valid():
        search_query = search_form.cleaned_data['s']
        search_model = search_form.cleaned_data['t']

        # get all codered models
        pagemodels = sorted(get_page_models(), key=lambda k: k.search_name)
        # get filterable models
        for model in pagemodels:
            if model.search_filterable:
                pagetypes.append(model)

        # get backend
        backend = get_search_backend()

        # DB search. Since this backend can't handle inheritance or scoring,
        # search specified page types in the desired order and chain the results together.
        # This provides better search results than simply searching limited fields on CoderedPage.
        db_models = []
        if backend.__class__ == db.SearchBackend:
            for model in get_page_models():
                if model.search_db_include:
                    db_models.append(model)
            db_models = sorted(db_models, reverse=True, key=lambda k: k.search_db_boost)

        if backend.__class__ == db.SearchBackend and db_models:
            for model in db_models:
                # if search_model is provided, only search on that model
                if not search_model or search_model == ContentType.objects.get_for_model(model).model:  # noqa
                    curr_results = model.objects.live().search(search_query)
                    if results:
                        results = list(chain(results, curr_results))
                    else:
                        results = curr_results

        # Fallback for any other search backend
        else:
            if search_model:
                try:
                    model = ContentType.objects.get(model=search_model).model_class()
                    results = model.objects.live().search(search_query)
                except ContentType.DoesNotExist:
                    results = None
            else:
                results = CoderedPage.objects.live().order_by('-last_published_at').search(search_query)  # noqa

        # paginate results
        if results:
            paginator = Paginator(results, GeneralSettings.for_request(request).search_num_results)
            page = request.GET.get('p', 1)
            try:
                results_paginated = paginator.page(page)
            except PageNotAnInteger:
                results_paginated = paginator.page(1)
            except EmptyPage:
                results_paginated = paginator.page(1)
            except InvalidPage:
                results_paginated = paginator.page(1)

        # Log the query so Wagtail can suggest promoted results
        Query.get(search_query).add_hit()

    # Render template
    return render(request, 'coderedcms/pages/search.html', {
        'request': request,
        'pagetypes': pagetypes,
        'form': search_form,
        'results': results,
        'results_paginated': results_paginated
    })


@login_required
def serve_protected_file(request, path):
    """
    Function that serves protected files uploaded from forms.
    """
    fullpath = os.path.join(cr_settings['PROTECTED_MEDIA_ROOT'], path)
    if os.path.isfile(fullpath):
        mimetype, encoding = mimetypes.guess_type(fullpath)
        with open(fullpath, 'rb') as f:
            response = HttpResponse(f.read(), content_type=mimetype)
        if encoding:
            response["Content-Encoding"] = encoding

        return response
    raise Http404()


def favicon(request):
    icon = LayoutSettings.for_request(request).favicon
    if icon:
        return HttpResponsePermanentRedirect(icon.get_rendition('original').url)
    raise Http404()


def robots(request):
    return render(
        request,
        'robots.txt',
        content_type='text/plain'
    )


def event_generate_single_ical_for_event(request):
    if request.method == "POST":
        event_pk = request.POST['event_pk']
        event_page_models = CoderedEventPage.__subclasses__()
        dt_start_str = utils.fix_ical_datetime_format(request.POST['datetime_start'])
        dt_end_str = utils.fix_ical_datetime_format(request.POST['datetime_end'])

        dt_start = datetime.strptime(dt_start_str, "%Y-%m-%dT%H:%M:%S%z") if dt_start_str else None
        dt_end = datetime.strptime(dt_end_str, "%Y-%m-%dT%H:%M:%S%z") if dt_end_str else None
        for event_page_model in event_page_models:
            try:
                event = event_page_model.objects.get(pk=event_pk)
                break
            except event_page_model.DoesNotExist:
                pass
        ical = Calendar()
        ical.add_component(event.create_single_ical(dt_start=dt_start, dt_end=dt_end))
        response = HttpResponse(ical.to_ical(), content_type="text/calendar")
        response['Filename'] = "{0}.ics".format(event.slug)
        response['Content-Disposition'] = 'attachment; filename={0}.ics'.format(event.slug)
        return response
    raise Http404()


def event_generate_recurring_ical_for_event(request):
    if request.method == "POST":
        event_pk = request.POST['event_pk']
        event_page_models = CoderedEventPage.__subclasses__()
        for event_page_model in event_page_models:
            try:
                event = event_page_model.objects.get(pk=event_pk)
                break
            except event_page_model.DoesNotExist:
                pass
        ical = Calendar()
        for e in event.create_recurring_ical():
            ical.add_component(e)
        response = HttpResponse(ical.to_ical(), content_type="text/calendar")
        response['Filename'] = "{0}.ics".format(event.slug)
        response['Content-Disposition'] = 'attachment; filename={0}.ics'.format(event.slug)
        return response
    raise Http404()


def event_generate_ical_for_calendar(request):
    if request.method == "POST":
        try:
            page = CoderedPage.objects.get(id=request.POST.get('page_id')).specific
            print(page)
        except ValueError:
            raise Http404

        ical = Calendar()
        for event_page in page.get_index_children():
            for e in event_page.specific.create_recurring_ical():
                ical.add_component(e)
        response = HttpResponse(ical.to_ical(), content_type="text/calendar")
        response['Filename'] = "calendar.ics"
        response['Content-Disposition'] = 'attachment; filename=calendar.ics'
        return response
    raise Http404()


def event_get_calendar_events(request):
    if request.is_ajax():
        try:
            page = CoderedPage.objects.get(id=request.GET.get('pid')).specific
        except ValueError:
            raise Http404
        start_str = request.GET.get('start')
        start = datetime.strptime(start_str[:10], "%Y-%m-%d") if start_str else None
        end_str = request.GET.get('end')
        end = datetime.strptime(end_str[:10], "%Y-%m-%d") if end_str else None
        return JsonResponse(page.get_calendar_events(start=start, end=end), safe=False)
    raise Http404()


@login_required
def import_pages_from_csv_file(request):
    """
    Overwrite of the `import_pages` view from wagtailimportexport.  By default, the `import_pages`
    view expects a json file to be uploaded.  This view converts the uploaded csv into the json
    format that the importer expects.
    """

    if request.method == 'POST':
        form = ImportPagesFromCSVFileForm(request.POST, request.FILES)
        if form.is_valid():
            import_data = convert_csv_to_json(
                form.cleaned_data['file'].read().decode('utf-8').splitlines(),
                form.cleaned_data['page_type']
            )
            parent_page = form.cleaned_data['parent_page']
            try:
                page_count = import_pages(import_data, parent_page)
            except LookupError as e:
                messages.error(request, _(
                    "Import failed: %(reason)s") % {'reason': e}
                )
            else:
                messages.success(request, ungettext(
                    "%(count)s page imported.",
                    "%(count)s pages imported.",
                    page_count) % {'count': page_count}
                )
            return redirect('wagtailadmin_explore', parent_page.pk)
    else:
        form = ImportPagesFromCSVFileForm()

    return render(request, 'wagtailimportexport/import_from_csv.html', {
        'form': form,
    })



import os
from setuptools import setup
from coderedcms import __version__

with open(os.path.join(os.path.dirname(__file__), "README.md"), encoding="utf8") as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='coderedcms',
    version=__version__,
    packages=['coderedcms'],
    include_package_data=True,
    license='BSD License',
    description='Wagtail-based CMS by CodeRed for building marketing websites.',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/coderedcorp/coderedcms',
    author='CodeRed LLC',
    author_email='info@coderedcorp.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Wagtail',
        'Framework :: Wagtail :: 2',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: Site Management',
    ],
    python_requires='>=3.6',
    install_requires=[
        'beautifulsoup4>=4.8,<4.10',    # should be the same as wagtail
        'django-eventtools==1.0.*',
        'django-bootstrap4>=2.0,<2.4',
        'Django>=2.2,<3.3',             # should be the same as wagtail
        'geocoder==1.38.*',
        'icalendar==4.0.*',
        'wagtail==2.13.*',
        'wagtailfontawesome>=1.2.*',
        'wagtail-cache==1.*',
        'wagtail-import-export>=0.2,<0.3',
        'wagtail-seo==1.*',
    ],
    entry_points={
        "console_scripts": [
            "coderedcms=coderedcms.bin.coderedcms:main"
        ]
    },
    zip_safe=False,
)



from modelcluster.fields import ParentalKey
from coderedcms.forms import CoderedFormField
from coderedcms.models import (
    CoderedArticlePage,
    CoderedArticleIndexPage,
    CoderedEventIndexPage,
    CoderedEventPage,
    CoderedEventOccurrence,
    CoderedEmail,
    CoderedFormPage,
    CoderedLocationIndexPage,
    CoderedLocationPage,
    CoderedStreamFormPage,
    CoderedWebPage
)


class ArticlePage(CoderedArticlePage):
    """
    Article, suitable for news or blog content.
    """
    class Meta:
        verbose_name = 'Article'
        ordering = ['-first_published_at', ]

    # Only allow this page to be created beneath an ArticleIndexPage.
    parent_page_types = ['testapp.ArticleIndexPage']

    template = 'coderedcms/pages/article_page.html'
    search_template = 'coderedcms/pages/article_page.search.html'


class ArticleIndexPage(CoderedArticleIndexPage):
    """
    Shows a list of article sub-pages.
    """
    class Meta:
        verbose_name = 'Article Landing Page'
    index_order_by_default = ''

    # Override to specify custom index ordering choice/default.
    index_query_pagemodel = 'testapp.ArticlePage'

    # Only allow ArticlePages beneath this page.
    subpage_types = ['testapp.ArticlePage']

    template = 'coderedcms/pages/article_index_page.html'


class FormPage(CoderedFormPage):
    """
    A page with an html <form>.
    """
    class Meta:
        verbose_name = 'Form'

    template = 'coderedcms/pages/form_page.html'


class FormPageField(CoderedFormField):
    """
    A field that links to a FormPage.
    """
    class Meta:
        ordering = ['sort_order']

    page = ParentalKey('FormPage', related_name='form_fields')


class FormConfirmEmail(CoderedEmail):
    """
    Sends a confirmation email after submitting a FormPage.
    """
    page = ParentalKey('FormPage', related_name='confirmation_emails')


class WebPage(CoderedWebPage):
    """
    General use page with featureful streamfield and SEO attributes.
    """
    class Meta:
        verbose_name = 'Web Page'

    template = 'coderedcms/pages/web_page.html'


class EventPage(CoderedEventPage):
    class Meta:
        verbose_name = 'Event Page'

    parent_page_types = ['testapp.EventIndexPage']
    subpage_types = []
    template = 'coderedcms/pages/event_page.html'


class EventIndexPage(CoderedEventIndexPage):
    """
    Shows a list of event sub-pages.
    """
    class Meta:
        verbose_name = 'Events Landing Page'

    index_query_pagemodel = 'testapp.EventPage'
    index_order_by_default = ''

    # Only allow EventPages beneath this page.
    subpage_types = ['testapp.EventPage']

    template = 'coderedcms/pages/event_index_page.html'


class EventOccurrence(CoderedEventOccurrence):
    event = ParentalKey(EventPage, related_name='occurrences')


class LocationPage(CoderedLocationPage):
    """
    A page that holds a location.  This could be a store, a restaurant, etc.
    """
    class Meta:
        verbose_name = 'Location Page'

    template = 'coderedcms/pages/location_page.html'

    # Only allow LocationIndexPages above this page.
    parent_page_types = ['testapp.LocationIndexPage']


class LocationIndexPage(CoderedLocationIndexPage):
    """
    A page that holds a list of locations and displays them with a Google Map.
    This does require a Google Maps API Key that can be defined in Settings > Google API Settings
    """
    class Meta:
        verbose_name = 'Location Landing Page'

    # Override to specify custom index ordering choice/default.
    index_query_pagemodel = 'testapp.LocationPage'

    # Only allow LocationPages beneath this page.
    subpage_types = ['testapp.LocationPage']

    template = 'coderedcms/pages/location_index_page.html'


class StreamFormPage(CoderedStreamFormPage):
    class Meta:
        verbose_name = 'Stream Form'

    template = 'coderedcms/pages/stream_form_page.html'


class StreamFormConfirmEmail(CoderedEmail):
    page = ParentalKey('StreamFormPage', related_name='confirmation_emails')



import csv
import copy

from django import forms
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import transaction

from wagtail.core.models import Page

from wagtailimportexport.forms import ImportFromFileForm
from wagtailimportexport.importing import update_page_references

from coderedcms.forms import get_page_model_choices


class ImportPagesFromCSVFileForm(ImportFromFileForm):
    page_type = forms.ChoiceField(choices=get_page_model_choices)


@transaction.atomic()
def import_pages(import_data, parent_page):
    """
    Overwrite of the wagtailimportexport `import_page` function to handle generic csvs.
    The standard `import_pages` assumes that your pages will have a pk from the exported
    json files.  It does not facilitate the idea that the pages you import will be
    new pages.
    """

    pages_by_original_id = {}

    # First create the base Page records; these contain no foreign keys, so this allows us to
    # build a complete mapping from old IDs to new IDs before we go on to importing the
    # specific page models, which may require us to rewrite page IDs within foreign keys / rich
    # text / streamfields.
    page_content_type = ContentType.objects.get_for_model(Page)

    for page_record in import_data['pages']:
        # build a base Page instance from the exported content
        # (so that we pick up its title and other core attributes)
        page = Page.from_serializable_data(page_record['content'])

        # clear id and treebeard-related fields so that
        # they get reassigned when we save via add_child
        page.id = None
        page.path = None
        page.depth = None
        page.numchild = 0
        page.url_path = None
        page.content_type = page_content_type
        parent_page.add_child(instance=page)

        # Custom Code to add the new pk back into the original page record.
        page_record['content']['pk'] = page.pk

        pages_by_original_id[page.id] = page

    for page_record in import_data['pages']:
        # Get the page model of the source page by app_label and model name
        # The content type ID of the source page is not in general the same
        # between the source and destination sites but the page model needs
        # to exist on both.
        # Raises LookupError exception if there is no matching model
        model = apps.get_model(page_record['app_label'], page_record['model'])

        specific_page = model.from_serializable_data(
            page_record['content'],
            check_fks=False,
            strict_fks=False
        )
        base_page = pages_by_original_id[specific_page.id]
        specific_page.page_ptr = base_page
        specific_page.__dict__.update(base_page.__dict__)
        specific_page.content_type = ContentType.objects.get_for_model(model)
        update_page_references(specific_page, pages_by_original_id)
        specific_page.save()

    return len(import_data['pages'])


def convert_csv_to_json(csv_file, page_type):
    pages_json = {"pages": []}
    default_page_data = {"app_label": "website", "content": {"pk": None}, "model": page_type}

    pages_csv_dict = csv.DictReader(csv_file)
    for row in pages_csv_dict:
        page_dict = copy.deepcopy(default_page_data)
        page_dict['content'].update(row)
        pages_json['pages'].append(page_dict)
    return pages_json



# Generated by Django 2.1.7 on 2019-03-22 17:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('coderedcms', '0010_remove_generalsettings_robots'),
        ('coderedcms', '0010_generalsettings_email'),
    ]

    operations = [
    ]



from django.test import Client
from wagtail.tests.utils import WagtailPageTests
from wagtail.core.models import Site

from coderedcms.tests.testapp.models import WebPage
from coderedcms.models.wagtailsettings_models import AnalyticsSettings


class AnalyticsSettingsTestCase(WagtailPageTests):
    """
    Test that the relevant analytics settings appear in the homepage HTML.
    """

    model = WebPage

    def setUp(self):
        # HTTP client.
        self.client = Client()

        # Use home page and default site.
        self.site = Site.objects.filter(is_default_site=True)[0]
        self.homepage = WebPage.objects.get(url_path="/home/")

        # Populate settings.
        self.settings = AnalyticsSettings.for_site(self.site)
        self.settings.ga_tracking_id = "UA-123"
        self.settings.head_scripts = "<script>evil_tracker</script>"
        self.settings.body_scripts = "<script>annoying_tracker</script>"
        self.settings.save()

    def test_get(self):
        """
        Tests to make sure the page serves a 200 from a GET request.
        """
        response = self.client.get(self.homepage.url, follow=True)
        self.assertEqual(response.status_code, 200)

    def test_ga_tracking_id(self):
        """
        Make sure the ga_tracking_id is present.
        """
        response = self.client.get(self.homepage.url, follow=True)
        self.assertIn(self.settings.ga_tracking_id, str(response.content), 1)

    def test_head_scripts(self):
        """
        Make sure the head_scripts is present.
        """
        response = self.client.get(self.homepage.url, follow=True)
        self.assertInHTML(self.settings.head_scripts, str(response.content), 1)

    def test_body_scripts(self):
        """
        Make sure the body_scripts is present.
        """
        response = self.client.get(self.homepage.url, follow=True)
        self.assertInHTML(self.settings.body_scripts, str(response.content), 1)



# Generated by Django 2.1.5 on 2019-01-08 19:41

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0040_page_draft_title'),
        ('coderedcms', '0005_auto_20181214_2214'),
    ]

    operations = [
        migrations.CreateModel(
            name='MailchimpApiSettings',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('mailchimp_api_key', models.CharField(blank=True, help_text='The API Key used for Mailchimp.', max_length=255, verbose_name='Mailchimp API Key')),
                ('site', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.Site')),
            ],
            options={
                'verbose_name': 'Mailchimp API Settings',
            },
        ),
    ]



"""
Django settings for {{ project_name }} project.

Generated by 'django-admin startproject' using Django {{ django_version }}.

For more information on this file, see
https://docs.djangoproject.com/en/{{ docs_version }}/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/{{ docs_version }}/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
from django.utils.translation import gettext_lazy as _  # noqa


PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BASE_DIR = os.path.dirname(PROJECT_DIR)


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/{{ docs_version }}/howto/deployment/checklist/


# Application definition

INSTALLED_APPS = [
    # Test
    'coderedcms.tests.testapp',

    # CodeRed CMS
    'coderedcms',
    'bootstrap4',
    'modelcluster',
    'taggit',
    'wagtailfontawesome',
    'wagtailcache',
    'wagtailimportexport',
    'wagtailseo',

    # Wagtail
    'wagtail.contrib.forms',
    'wagtail.contrib.redirects',
    'wagtail.embeds',
    'wagtail.sites',
    'wagtail.users',
    'wagtail.snippets',
    'wagtail.documents',
    'wagtail.images',
    'wagtail.search',
    'wagtail.core',
    'wagtail.contrib.settings',
    'wagtail.contrib.modeladmin',
    'wagtail.contrib.table_block',
    'wagtail.admin',

    # Django
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    "django.contrib.sitemaps",
]

MIDDLEWARE = [
    # Save pages to cache. Must be FIRST.
    'wagtailcache.cache.UpdateCacheMiddleware',

    # Common functionality
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.common.CommonMiddleware',

    # Security
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',

    # Error reporting. Uncomment this to receive emails when a 404 is triggered.
    # 'django.middleware.common.BrokenLinkEmailsMiddleware',

    # CMS functionality
    'wagtail.contrib.redirects.middleware.RedirectMiddleware',

    # Fetch from cache. Must be LAST.
    'wagtailcache.cache.FetchFromCacheMiddleware',
]

ROOT_URLCONF = 'coderedcms.tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'wagtail.contrib.settings.context_processors.settings',
            ],
        },
    },
]


# Database
# https://docs.djangoproject.com/en/{{ docs_version }}/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        'TEST': {
            'NAME': os.path.join(BASE_DIR, 'test_db.sqlite3'),
        }
    }
}


# Password validation
# https://docs.djangoproject.com/en/{{ docs_version }}/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/{{ docs_version }}/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'America/New_York'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/{{ docs_version }}/howto/static-files/

STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
]

STATIC_ROOT = os.path.join(BASE_DIR, 'static')
STATIC_URL = '/teststatic/'

MEDIA_ROOT = os.path.join(BASE_DIR, 'media')
MEDIA_URL = '/testmedia/'


# Login

LOGIN_URL = 'wagtailadmin_login'
LOGIN_REDIRECT_URL = 'wagtailadmin_home'


# Wagtail settings

WAGTAIL_SITE_NAME = ""

WAGTAIL_ENABLE_UPDATE_CHECK = False

# Base URL to use when referring to full URLs within the Wagtail admin backend -
# e.g. in notification emails. Don't include '/admin' or a trailing slash
BASE_URL = ''


# Bootstrap

BOOTSTRAP4 = {
    # set to blank since coderedcms already loads jquery and bootstrap
    'jquery_url': '',
    'base_url': '',
    # remove green highlight on inputs
    'success_css_class': ''
}


# Tags

TAGGIT_CASE_INSENSITIVE = True

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

WAGTAIL_CACHE = False

SECRET_KEY = 'not needed'

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

NOSE_ARGS = ['--nocapture',
             '--nologcapture', ]
